# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from ...utils import is_build_mode
from ..core import DATAFRAME_TYPE
from ..utils import wrap_notimplemented_exception
from ..ufunc.tensor import register_tensor_ufunc
from .abs import abs_, DataFrameAbs
from .add import add, radd, DataFrameAdd
from .bitwise_and import bitand, rbitand, DataFrameAnd
from .bitwise_or import bitor, rbitor, DataFrameOr
from .bitwise_xor import bitxor, rbitxor, DataFrameXor
from .subtract import subtract, rsubtract, DataFrameSubtract
from .multiply import mul, rmul, DataFrameMul
from .floordiv import floordiv, rfloordiv, DataFrameFloorDiv
from .truediv import truediv, rtruediv, DataFrameTrueDiv
from .mod import mod, rmod, DataFrameMod
from .power import power, rpower, DataFramePower
from .equal import eq, DataFrameEqual
from .not_equal import ne, DataFrameNotEqual
from .less import lt, DataFrameLess
from .greater import gt, DataFrameGreater
from .less_equal import le, DataFrameLessEqual
from .greater_equal import ge, DataFrameGreaterEqual
from .invert import invert, DataFrameNot
from .is_ufuncs import DataFrameIsNan, DataFrameIsInf, DataFrameIsFinite
from .negative import DataFrameNegative, negative
from .log import DataFrameLog
from .log2 import DataFrameLog2
from .log10 import DataFrameLog10
from .sin import DataFrameSin
from .cos import DataFrameCos
from .tan import DataFrameTan
from .sinh import DataFrameSinh
from .cosh import DataFrameCosh
from .tanh import DataFrameTanh
from .arcsin import DataFrameArcsin
from .arccos import DataFrameArccos
from .arctan import DataFrameArctan
from .arcsinh import DataFrameArcsinh
from .arccosh import DataFrameArccosh
from .arctanh import DataFrameArctanh
from .radians import DataFrameRadians
from .degrees import DataFrameDegrees
from .ceil import DataFrameCeil
from .floor import DataFrameFloor
from .around import DataFrameAround, around
from .sqrt import DataFrameSqrt
from .exp import DataFrameExp
from .exp2 import DataFrameExp2
from .expm1 import DataFrameExpm1
from .dot import dot


def _wrap_eq():
    @functools.wraps(eq)
    def call(df, other, **kw):
        if is_build_mode():
            return df._equals(other)
        return _wrap_comparison(eq)(df, other, **kw)
    return call


def _wrap_comparison(func):
    @functools.wraps(func)
    def call(df, other, **kw):
        if isinstance(df, DATAFRAME_TYPE) and isinstance(other, DATAFRAME_TYPE):
            # index and columns should be identical
            for index_type in ['index_value', 'columns_value']:
                left, right = getattr(df, index_type), getattr(other, index_type)
                if left.has_value() and right.has_value():
                    # if df and other's index or columns has value
                    index_eq = left.to_pandas().equals(right.to_pandas())
                else:
                    index_eq = left.key == right.key
                if not index_eq:
                    raise ValueError('Can only compare '
                                     'identically-labeled DataFrame object')
        return wrap_notimplemented_exception(func)(df, other, **kw)
    return call


def _install():
    from ..core import DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE

    def _register_method(cls, name, func, wrapper=None):
        if wrapper is None:
            @functools.wraps(func)
            def wrapper(df, *args, **kwargs):
                return func(df, *args, **kwargs)

        try:
            if issubclass(cls, DATAFRAME_TYPE):
                wrapper.__doc__ = func.__frame_doc__
            elif issubclass(cls, SERIES_TYPE):
                wrapper.__doc__ = func.__series_doc__
            else:
                wrapper = func
        except AttributeError:
            wrapper = func

        wrapper.__name__ = func.__name__
        setattr(cls, name, wrapper)

    def _register_bin_method(cls, name, func):
        def call_df_fill(df, other, axis='columns', level=None, fill_value=None):
            return func(df, other, axis=axis, level=level, fill_value=fill_value)

        def call_df_no_fill(df, other, axis='columns', level=None):
            return func(df, other, axis=axis, level=level)

        def call_series_fill(df, other, level=None, fill_value=None, axis=0):
            return func(df, other, axis=axis, level=level, fill_value=fill_value)

        def call_series_no_fill(df, other, level=None, axis=0):
            return func(df, other, axis=axis, level=level)

        if issubclass(cls, DATAFRAME_TYPE):
            call = call_df_fill if 'fill_value' in func.__code__.co_varnames else call_df_no_fill
        elif issubclass(cls, SERIES_TYPE):
            call = call_series_fill if 'fill_value' in func.__code__.co_varnames else call_series_no_fill
        else:
            call = None
        return _register_method(cls, name, func, wrapper=call)

    # register mars tensor ufuncs
    ufunc_ops = [
        # unary
        DataFrameAbs, DataFrameLog, DataFrameLog2, DataFrameLog10,
        DataFrameSin, DataFrameCos, DataFrameTan,
        DataFrameSinh, DataFrameCosh, DataFrameTanh,
        DataFrameArcsin, DataFrameArccos, DataFrameArctan,
        DataFrameArcsinh, DataFrameArccosh, DataFrameArctanh,
        DataFrameRadians, DataFrameDegrees,
        DataFrameCeil, DataFrameFloor, DataFrameAround,
        DataFrameExp, DataFrameExp2, DataFrameExpm1,
        DataFrameSqrt, DataFrameNot, DataFrameIsNan,
        DataFrameIsInf, DataFrameIsFinite, DataFrameNegative,
        # binary
        DataFrameAdd, DataFrameEqual, DataFrameFloorDiv,
        DataFrameGreater, DataFrameGreaterEqual, DataFrameLess,
        DataFrameLessEqual, DataFrameAnd, DataFrameOr, DataFrameXor,
        DataFrameMod, DataFrameMul, DataFrameNotEqual, DataFramePower,
        DataFrameSubtract, DataFrameTrueDiv,
    ]
    for ufunc_op in ufunc_ops:
        register_tensor_ufunc(ufunc_op)

    for entity in DATAFRAME_TYPE + SERIES_TYPE:
        setattr(entity, '__abs__', abs_)
        setattr(entity, 'abs', abs_)
        _register_method(entity, 'round', around)
        setattr(entity, '__invert__', invert)

        setattr(entity, '__add__', wrap_notimplemented_exception(add))
        setattr(entity, '__radd__', wrap_notimplemented_exception(radd))
        _register_bin_method(entity, 'add', add)
        _register_bin_method(entity, 'radd', radd)

        setattr(entity, '__sub__', wrap_notimplemented_exception(subtract))
        setattr(entity, '__rsub__', wrap_notimplemented_exception(rsubtract))
        _register_bin_method(entity, 'sub', subtract)
        _register_bin_method(entity, 'rsub', rsubtract)

        setattr(entity, '__mul__', wrap_notimplemented_exception(mul))
        setattr(entity, '__rmul__', wrap_notimplemented_exception(rmul))
        _register_bin_method(entity, 'mul', mul)
        _register_bin_method(entity, 'multiply', mul)
        _register_bin_method(entity, 'rmul', rmul)

        setattr(entity, '__floordiv__', wrap_notimplemented_exception(floordiv))
        setattr(entity, '__rfloordiv__', wrap_notimplemented_exception(rfloordiv))
        setattr(entity, '__truediv__', wrap_notimplemented_exception(truediv))
        setattr(entity, '__rtruediv__', wrap_notimplemented_exception(rtruediv))
        setattr(entity, '__div__', wrap_notimplemented_exception(truediv))
        setattr(entity, '__rdiv__', wrap_notimplemented_exception(rtruediv))
        _register_bin_method(entity, 'floordiv', floordiv)
        _register_bin_method(entity, 'rfloordiv', rfloordiv)
        _register_bin_method(entity, 'truediv', truediv)
        _register_bin_method(entity, 'rtruediv', rtruediv)
        _register_bin_method(entity, 'div', truediv)
        _register_bin_method(entity, 'rdiv', rtruediv)

        setattr(entity, '__mod__', wrap_notimplemented_exception(mod))
        setattr(entity, '__rmod__', wrap_notimplemented_exception(rmod))
        _register_bin_method(entity, 'mod', mod)
        _register_bin_method(entity, 'rmod', rmod)

        setattr(entity, '__pow__', wrap_notimplemented_exception(power))
        setattr(entity, '__rpow__', wrap_notimplemented_exception(rpower))
        _register_bin_method(entity, 'pow', power)
        _register_bin_method(entity, 'rpow', rpower)

        setattr(entity, '__eq__', _wrap_eq())
        setattr(entity, '__ne__', _wrap_comparison(ne))
        setattr(entity, '__lt__', _wrap_comparison(lt))
        setattr(entity, '__gt__', _wrap_comparison(gt))
        setattr(entity, '__ge__', _wrap_comparison(ge))
        setattr(entity, '__le__', _wrap_comparison(le))
        _register_bin_method(entity, 'eq', eq)
        _register_bin_method(entity, 'ne', ne)
        _register_bin_method(entity, 'lt', lt)
        _register_bin_method(entity, 'gt', gt)
        _register_bin_method(entity, 'ge', ge)
        _register_bin_method(entity, 'le', le)

        setattr(entity, '__matmul__', dot)
        _register_method(entity, 'dot', dot)

        setattr(entity, '__and__', wrap_notimplemented_exception(bitand))
        setattr(entity, '__rand__', wrap_notimplemented_exception(rbitand))

        setattr(entity, '__or__', wrap_notimplemented_exception(bitor))
        setattr(entity, '__ror__', wrap_notimplemented_exception(rbitor))

        setattr(entity, '__xor__', wrap_notimplemented_exception(bitxor))
        setattr(entity, '__rxor__', wrap_notimplemented_exception(rbitxor))

        setattr(entity, '__neg__', wrap_notimplemented_exception(negative))

    for entity in INDEX_TYPE:
        setattr(entity, '__eq__', _wrap_eq())


_install()
del _install



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np

from ... import opcodes as OperandDef
from ...core import Base, Entity, OutputType
from ...serialize import ValueType, ListField, StringField, BoolField, AnyField
from ...tiles import TilesError
from ...utils import lazy_import, check_chunks_unknown_shape
from ..utils import parse_index, build_empty_df, build_empty_series, \
    standardize_range_index, validate_axis
from ..operands import DataFrameOperand, DataFrameOperandMixin, SERIES_TYPE

cudf = lazy_import('cudf', globals=globals())


class DataFrameConcat(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.CONCATENATE

    _axis = AnyField('axis')
    _join = StringField('join')
    _ignore_index = BoolField('ignore_index')
    _keys = ListField('keys')
    _levels = ListField('levels')
    _names = ListField('names')
    _verify_integrity = BoolField('verify_integrity')
    _sort = BoolField('sort')
    _copy = BoolField('copy')

    def __init__(self, axis=None, join=None, ignore_index=None,
                 keys=None, levels=None, names=None, verify_integrity=None,
                 sort=None, copy=None, sparse=None, output_types=None, **kw):
        super().__init__(
            _axis=axis, _join=join, _ignore_index=ignore_index,
            _keys=keys, _levels=levels, _names=names,
            _verify_integrity=verify_integrity, _sort=sort, _copy=copy,
            _sparse=sparse, _output_types=output_types, **kw)

    @property
    def axis(self):
        return self._axis

    @property
    def join(self):
        return self._join

    @property
    def ignore_index(self):
        return self._ignore_index

    @property
    def keys(self):
        return self._keys

    @property
    def level(self):
        return self._levels

    @property
    def name(self):
        return self._names

    @property
    def verify_integrity(self):
        return self._verify_integrity

    @property
    def sort(self):
        return self._sort

    @property
    def copy_(self):
        return self._copy

    @classmethod
    def _tile_dataframe(cls, op):
        from ..indexing.iloc import DataFrameIlocGetItem

        out_df = op.outputs[0]
        inputs = op.inputs
        axis = op.axis

        if not all(inputs[i].nsplits[1 - axis] == inputs[i + 1].nsplits[1 - axis]
                   for i in range(len(inputs) - 1)):
            # need rechunk
            check_chunks_unknown_shape(inputs, TilesError)
            normalized_nsplits = {1 - axis: inputs[0].nsplits[1 - axis]}
            inputs = [inp.rechunk(normalized_nsplits)._inplace_tile()
                      for inp in inputs]

        out_chunks = []
        nsplits = []
        cum_index = 0
        for df in inputs:
            for c in df.chunks:
                if op.axis == 0:
                    index = (c.index[0] + cum_index, c.index[1])
                else:
                    index = (c.index[0], c.index[1] + cum_index)

                iloc_op = DataFrameIlocGetItem(indexes=(slice(None),) * 2)
                out_chunks.append(iloc_op.new_chunk([c], shape=c.shape, index=index,
                                                    dtypes=c.dtypes,
                                                    index_value=c.index_value,
                                                    columns_value=c.columns_value))
            nsplits.extend(df.nsplits[op.axis])
            cum_index += len(df.nsplits[op.axis])
        out_nsplits = (tuple(nsplits), inputs[0].nsplits[1]) \
            if op.axis == 0 else (inputs[0].nsplits[0], tuple(nsplits))

        if op.ignore_index:
            out_chunks = standardize_range_index(out_chunks)

        new_op = op.copy()
        return new_op.new_dataframes(op.inputs, out_df.shape,
                                     nsplits=out_nsplits, chunks=out_chunks,
                                     dtypes=out_df.dtypes,
                                     index_value=out_df.index_value,
                                     columns_value=out_df.columns_value)

    @classmethod
    def _tile_series(cls, op):
        from ..indexing.iloc import SeriesIlocGetItem

        out = op.outputs[0]
        inputs = op.inputs
        out_chunks = []

        if op.axis == 1:
            check_chunks_unknown_shape(inputs, TilesError)
            inputs = [item.rechunk(op.inputs[0].nsplits)._inplace_tile() for item in inputs]

        cum_index = 0
        nsplits = []
        for series in inputs:
            for c in series.chunks:
                if op.axis == 0:
                    index = (c.index[0] + cum_index,)
                    shape = c.shape
                else:
                    index = (c.index[0], cum_index)
                    shape = (c.shape[0], 1)
                iloc_op = SeriesIlocGetItem(indexes=(slice(None),))
                out_chunks.append(iloc_op.new_chunk([c], shape=shape, index=index,
                                                    index_value=c.index_value,
                                                    dtype=c.dtype,
                                                    name=c.name))
            if op.axis == 0:
                nsplits.extend(series.nsplits[0])
                cum_index += len(series.nsplits[op.axis])
            else:
                nsplits.append(1)
                cum_index += 1

        if op.ignore_index:
            out_chunks = standardize_range_index(out_chunks)

        new_op = op.copy()
        if op.axis == 0:
            nsplits = (tuple(nsplits),)
            return new_op.new_seriess(op.inputs, out.shape,
                                      nsplits=nsplits, chunks=out_chunks,
                                      dtype=out.dtype,
                                      index_value=out.index_value,
                                      name=out.name)
        else:
            nsplits = (inputs[0].nsplits[0], tuple(nsplits))
            return new_op.new_dataframes(op.inputs, out.shape,
                                         nsplits=nsplits, chunks=out_chunks,
                                         dtypes=out.dtypes,
                                         index_value=out.index_value,
                                         columns_value=out.columns_value)

    @classmethod
    def tile(cls, op):
        if isinstance(op.inputs[0], SERIES_TYPE):
            return cls._tile_series(op)
        else:
            return cls._tile_dataframe(op)

    @classmethod
    def execute(cls, ctx, op):
        def _base_concat(chunk, inputs):
            # auto generated concat when executing a DataFrame, Series or Index
            if chunk.op.output_types[0] == OutputType.dataframe:
                return _auto_concat_dataframe_chunks(chunk, inputs)
            elif chunk.op.output_types[0] == OutputType.series:
                return _auto_concat_series_chunks(chunk, inputs)
            elif chunk.op.output_types[0] == OutputType.index:
                return _auto_concat_index_chunks(chunk, inputs)
            elif chunk.op.output_types[0] == OutputType.categorical:
                return _auto_concat_categorical_chunks(chunk, inputs)
            else:  # pragma: no cover
                raise TypeError('Only DataFrameChunk, SeriesChunk, IndexChunk, '
                                'and CategoricalChunk can be automatically concatenated')

        def _auto_concat_dataframe_chunks(chunk, inputs):
            xdf = pd if isinstance(inputs[0], (pd.DataFrame, pd.Series)) or cudf is None else cudf

            if chunk.op.axis is not None:
                return xdf.concat(inputs, axis=op.axis)

            # auto generated concat when executing a DataFrame
            if len(inputs) == 1:
                ret = inputs[0]
            else:
                n_rows = len(set(inp.index[0] for inp in chunk.inputs))
                n_cols = int(len(inputs) // n_rows)
                assert n_rows * n_cols == len(inputs)

                concats = []
                for i in range(n_rows):
                    if n_cols == 1:
                        concats.append(inputs[i])
                    else:
                        concat = xdf.concat([inputs[i * n_cols + j] for j in range(n_cols)], axis=1)
                        concats.append(concat)

                if xdf is pd:
                    # The `sort=False` is to suppress a `FutureWarning` of pandas,
                    # when the index or column of chunks to concatenate is not aligned,
                    # which may happens for certain ops.
                    #
                    # See also Note [Columns of Left Join] in test_merge_execution.py.
                    ret = xdf.concat(concats, sort=False)
                else:
                    ret = xdf.concat(concats)
                    # cuDF will lost index name when concat two seriess.
                    ret.index.name = concats[0].index.name

            if getattr(chunk.index_value, 'should_be_monotonic', False):
                ret.sort_index(inplace=True)
            if getattr(chunk.columns_value, 'should_be_monotonic', False):
                ret.sort_index(axis=1, inplace=True)
            return ret

        def _auto_concat_series_chunks(chunk, inputs):
            # auto generated concat when executing a Series
            if len(inputs) == 1:
                concat = inputs[0]
            else:
                xdf = pd if isinstance(inputs[0], pd.Series) or cudf is None else cudf
                if chunk.op.axis is not None:
                    concat = xdf.concat(inputs, axis=chunk.op.axis)
                else:
                    concat = xdf.concat(inputs)
            if getattr(chunk.index_value, 'should_be_monotonic', False):
                concat.sort_index(inplace=True)
            return concat

        def _auto_concat_index_chunks(chunk, inputs):
            if len(inputs) == 1:
                xdf = pd if isinstance(inputs[0], pd.Index) or cudf is None else cudf
                concat_df = xdf.DataFrame(index=inputs[0])
            else:
                xdf = pd if isinstance(inputs[0], pd.Index) or cudf is None else cudf
                empty_dfs = [xdf.DataFrame(index=inp) for inp in inputs]
                concat_df = xdf.concat(empty_dfs, axis=0)
            if getattr(chunk.index_value, 'should_be_monotonic', False):
                concat_df.sort_index(inplace=True)
            return concat_df.index

        def _auto_concat_categorical_chunks(_, inputs):
            if len(inputs) == 1:  # pragma: no cover
                return inputs[0]
            else:
                # convert categorical into array
                arrays = [np.asarray(inp) for inp in inputs]
                array = np.concatenate(arrays)
                return pd.Categorical(array, categories=inputs[0].categories,
                                      ordered=inputs[0].ordered)

        chunk = op.outputs[0]
        inputs = [ctx[input.key] for input in op.inputs]

        if isinstance(inputs[0], tuple):
            ctx[chunk.key] = tuple(_base_concat(chunk, [input[i] for input in inputs])
                                   for i in range(len(inputs[0])))
        else:
            ctx[chunk.key] = _base_concat(chunk, inputs)

    @classmethod
    def _concat_index(cls, prev_index: pd.Index, cur_index: pd.Index):
        if isinstance(prev_index, pd.RangeIndex) and \
                isinstance(cur_index, pd.RangeIndex):
            # handle RangeIndex that append may generate huge amount of data
            # e.g. pd.RangeIndex(10_000) and pd.RangeIndex(10_000)
            # will generate a Int64Index full of data
            # for details see GH#1647
            prev_stop = prev_index.start + prev_index.size * prev_index.step
            cur_start = cur_index.start
            if prev_stop == cur_start and prev_index.step == cur_index.step:
                # continuous RangeIndex, still return RangeIndex
                return prev_index.append(cur_index)
            else:
                # otherwise, return an empty index
                return pd.Index([], dtype=prev_index.dtype)
        elif isinstance(prev_index, pd.RangeIndex):
            return pd.Index([], prev_index.dtype).append(cur_index)
        elif isinstance(cur_index, pd.RangeIndex):
            return prev_index.append(pd.Index([], cur_index.dtype))
        return prev_index.append(cur_index)

    def _call_series(self, objs):
        if self.axis == 0:
            row_length = 0
            index = None
            for series in objs:
                if index is None:
                    index = series.index_value.to_pandas()
                else:
                    index = self._concat_index(index, series.index_value.to_pandas())
                row_length += series.shape[0]
            if self.ignore_index:  # pragma: no cover
                index_value = parse_index(pd.RangeIndex(row_length))
            else:
                index_value = parse_index(index, objs)
            return self.new_series(objs, shape=(row_length,), dtype=objs[0].dtype,
                                   index_value=index_value, name=objs[0].name)
        else:
            col_length = 0
            columns = []
            dtypes = dict()
            undefined_name = 0
            for series in objs:
                if series.name is None:
                    dtypes[undefined_name] = series.dtype
                    undefined_name += 1
                    columns.append(undefined_name)
                else:
                    dtypes[series.name] = series.dtype
                    columns.append(series.name)
                col_length += 1
            if self.ignore_index or undefined_name == len(objs):
                columns_value = parse_index(pd.RangeIndex(col_length))
            else:
                columns_value = parse_index(pd.Index(columns), store_data=True)

            shape = (objs[0].shape[0], col_length)
            return self.new_dataframe(objs, shape=shape, dtypes=pd.Series(dtypes),
                                      index_value=objs[0].index_value,
                                      columns_value=columns_value)

    def _call_dataframes(self, objs):
        if self.axis == 0:
            row_length = 0
            index = None
            empty_dfs = []
            for df in objs:
                if index is None:
                    index = df.index_value.to_pandas()
                else:
                    index = self._concat_index(index, df.index_value.to_pandas())
                row_length += df.shape[0]
                if df.ndim == 2:
                    empty_dfs.append(build_empty_df(df.dtypes))
                else:
                    empty_dfs.append(build_empty_series(df.dtype, name=df.name))

            emtpy_result = pd.concat(empty_dfs, join=self.join, sort=True)
            shape = (row_length, emtpy_result.shape[1])
            columns_value = parse_index(emtpy_result.columns, store_data=True)

            if self.join == 'inner':
                objs = [o[list(emtpy_result.columns)] for o in objs]

            if self.ignore_index:  # pragma: no cover
                index_value = parse_index(pd.RangeIndex(row_length))
            else:
                index_value = parse_index(index, objs)

            new_objs = []
            for obj in objs:
                if obj.ndim != 2:
                    # series
                    new_obj = obj.to_frame().reindex(columns=emtpy_result.dtypes.index)
                else:
                    # dataframe
                    if list(obj.dtypes.index) != list(emtpy_result.dtypes.index):
                        new_obj = obj.reindex(columns=emtpy_result.dtypes.index)
                    else:
                        new_obj = obj
                new_objs.append(new_obj)

            return self.new_dataframe(new_objs, shape=shape, dtypes=emtpy_result.dtypes,
                                      index_value=index_value, columns_value=columns_value)
        else:
            col_length = 0
            empty_dfs = []
            for df in objs:
                if df.ndim == 2:
                    # DataFrame
                    col_length += df.shape[1]
                    empty_dfs.append(build_empty_df(df.dtypes))
                else:
                    # Series
                    col_length += 1
                    empty_dfs.append(build_empty_series(df.dtype, name=df.name))

            emtpy_result = pd.concat(empty_dfs, join=self.join, axis=1, sort=True)
            if self.ignore_index:
                columns_value = parse_index(pd.RangeIndex(col_length))
            else:
                columns_value = parse_index(pd.Index(emtpy_result.columns), store_data=True)

            if self.ignore_index or len({o.index_value.key for o in objs}) == 1:
                new_objs = [obj if obj.ndim == 2 else obj.to_frame()
                            for obj in objs]
            else:  # pragma: no cover
                raise NotImplementedError('Does not support concat dataframes '
                                          'which has different index')

            shape = (objs[0].shape[0], col_length)
            return self.new_dataframe(new_objs, shape=shape, dtypes=emtpy_result.dtypes,
                                      index_value=objs[0].index_value,
                                      columns_value=columns_value)

    def __call__(self, objs):
        if all(isinstance(obj, SERIES_TYPE) for obj in objs):
            self.output_types = [OutputType.series]
            return self._call_series(objs)
        else:
            self.output_types = [OutputType.dataframe]
            return self._call_dataframes(objs)


class GroupByConcat(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.GROUPBY_CONCAT

    _groups = ListField('groups', ValueType.key)
    _groupby_params = AnyField('groupby_params')

    def __init__(self, groups=None, groupby_params=None, output_types=None, **kw):
        super().__init__(_groups=groups, _groupby_params=groupby_params,
                         _output_types=output_types, **kw)

    @property
    def groups(self):
        return self._groups

    @property
    def groupby_params(self):
        return self._groupby_params

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)

        new_groups = []
        for _ in self._groups:
            new_groups.append(next(inputs_iter))
        self._groups = new_groups

        if isinstance(self._groupby_params['by'], list):
            by = []
            for v in self._groupby_params['by']:
                if isinstance(v, (Base, Entity)):
                    by.append(next(inputs_iter))
                else:
                    by.append(v)
            self._groupby_params['by'] = by

    @classmethod
    def execute(cls, ctx, op):
        input_data = [ctx[input.key] for input in op.groups]
        obj = pd.concat([d.obj for d in input_data])

        params = op.groupby_params.copy()
        if isinstance(params['by'], list):
            by = []
            for v in params['by']:
                if isinstance(v, Base):
                    by.append(ctx[v.key])
                else:
                    by.append(v)
            params['by'] = by
        selection = params.pop('selection', None)

        result = obj.groupby(**params)
        if selection:
            result = result[selection]

        ctx[op.outputs[0].key] = result


def concat(objs, axis=0, join='outer', ignore_index=False, keys=None, levels=None, names=None,
           verify_integrity=False, sort=False, copy=True):
    if not isinstance(objs, (list, tuple)):  # pragma: no cover
        raise TypeError('first argument must be an iterable of dataframe or series objects')
    axis = validate_axis(axis)
    if isinstance(objs, dict):  # pragma: no cover
        keys = objs.keys()
        objs = objs.values()
    if axis == 1 and join == 'inner':  # pragma: no cover
        raise NotImplementedError('inner join is not support when specify `axis=1`')
    if verify_integrity or sort or keys:  # pragma: no cover
        raise NotImplementedError('verify_integrity, sort, keys arguments are not supported now')
    op = DataFrameConcat(axis=axis, join=join, ignore_index=ignore_index, keys=keys, levels=levels,
                         names=names, verify_integrity=verify_integrity, sort=sort, copy=copy)

    return op(objs)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import time
from collections import defaultdict

import numpy as np

from .kvstore import KVStoreActor
from .session import SessionManagerActor, SessionActor
from .utils import SchedulerActor
from ..config import options
from ..utils import BlacklistSet

logger = logging.getLogger(__name__)


class ResourceHeartbeatActor(SchedulerActor):
    def __init__(self, resource_ref):
        super().__init__()
        self._resource_ref = resource_ref

    def post_create(self):
        super().post_create()
        self._resource_ref = self.ctx.actor_ref(self._resource_ref)
        self.ref().do_heartbeat(_tell=True)

    def do_heartbeat(self):
        self._resource_ref.heartbeat(_tell=True)
        self.ref().do_heartbeat(_tell=True, _delay=1)


class ResourceActor(SchedulerActor):
    """
    Actor managing free resources on workers
    """
    def __init__(self):
        super().__init__()
        self._meta_cache = dict()
        self._worker_blacklist = BlacklistSet(options.scheduler.worker_blacklist_time)
        self._worker_allocations = dict()

        self._last_heartbeat_time = time.time()
        self._last_heartbeat_interval = 0

        self._kv_store_ref = None
        self._heartbeat_ref = None

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        super().post_create()
        self.set_cluster_info_ref()

        self._kv_store_ref = self.ctx.actor_ref(KVStoreActor.default_uid())
        if not self.ctx.has_actor(self._kv_store_ref):
            self._kv_store_ref = None

        try:
            # we assign the heartbeat actor into another process
            # so it can sense whether the process is busy
            heartbeat_uid = self.ctx.distributor.make_same_process(
                ResourceHeartbeatActor.default_uid(), self.uid, delta=1)
        except AttributeError:
            heartbeat_uid = ResourceHeartbeatActor.default_uid()
        self._heartbeat_ref = self.ctx.create_actor(
            ResourceHeartbeatActor, self.ref(), uid=heartbeat_uid)

        self.ref().detect_dead_workers(_tell=True)

    def pre_destroy(self):
        self._heartbeat_ref.destroy()
        self.unset_cluster_info_ref()
        super().pre_destroy()

    def heartbeat(self):
        t = time.time()
        self._last_heartbeat_interval = t - self._last_heartbeat_time
        self._last_heartbeat_time = t

    def detect_dead_workers(self):
        """
        Remove worker when it does not update its status for a long time
        """
        # take latency of scheduler into consideration by
        # include interval of last heartbeat
        timeout = options.scheduler.status_timeout + self._last_heartbeat_interval
        dead_workers = []

        check_time = self._last_heartbeat_time
        for worker in list(self._meta_cache.keys()):
            worker_meta = self._meta_cache[worker]
            try:
                if check_time - worker_meta['update_time'] > timeout:
                    dead_workers.append(worker)
            except (KeyError, TypeError, ValueError):
                pass

        self.detach_dead_workers(dead_workers)
        self.ref().detect_dead_workers(_tell=True, _delay=1)

    def mark_workers_alive(self, workers):
        t = time.time()
        for w in workers:
            try:
                self._meta_cache[w]['update_time'] = t
            except KeyError:  # pragma: no cover
                pass

    def detach_dead_workers(self, workers):
        from ..worker.execution import ExecutionActor

        workers = [w for w in workers if w in self._meta_cache and
                   w not in self._worker_blacklist]

        if not workers:
            return

        logger.warning('Workers %r dead, detaching from ResourceActor.', workers)
        for w in workers:
            del self._meta_cache[w]
            self._worker_blacklist.add(w)

        self._broadcast_sessions(SessionActor.handle_worker_change, [], workers)
        self._broadcast_workers(ExecutionActor.handle_worker_change, [], workers)

    def get_worker_count(self):
        return len(self._meta_cache)

    def get_total_cpu_count(self):
        """Get all workers' cpu count"""
        return sum(v['hardware']['cpu_total'] for k, v in self._meta_cache.items()
                   if k not in self._worker_blacklist)

    def get_workers_meta(self):
        return dict((k, v) for k, v in self._meta_cache.items()
                    if k not in self._worker_blacklist)

    def get_worker_endpoints(self):
        return [k for k in self._meta_cache if k not in self._worker_blacklist]

    def get_worker_metas(self):  # pragma: no cover
        return {k: meta for k, meta in self._meta_cache.items() if k not in self._worker_blacklist}

    def select_worker(self, size=None):
        # NB: randomly choice the indices, rather than the worker list to avoid the `str` to `np.str_` casting.
        available_workers = [k for k in self._meta_cache if k not in self._worker_blacklist]
        idx = np.random.choice(range(len(available_workers)), size=size)
        if size is None:
            return available_workers[idx]
        else:
            return [available_workers[i] for i in idx]

    def set_worker_meta(self, worker, worker_meta):
        from ..worker.execution import ExecutionActor

        if worker in self._worker_blacklist:
            return

        is_new = worker not in self._meta_cache

        if worker_meta.get('update_time', 0) < self._meta_cache.get(worker, {}).get('update_time', 0):
            return

        self._meta_cache[worker] = worker_meta
        if self._kv_store_ref is not None:
            self._kv_store_ref.write(f'/workers/meta/{worker}', json.dumps(worker_meta),
                                     _tell=True, _wait=False)
            self._kv_store_ref.write('/workers/meta_timestamp', str(int(time.time())),
                                     _tell=True, _wait=False)
        if is_new:
            self._broadcast_sessions(SessionActor.handle_worker_change, [worker], [])
            self._broadcast_workers(ExecutionActor.handle_worker_change, [worker], [])

    def _broadcast_sessions(self, handler, *args, **kwargs):
        if hasattr(handler, '__name__'):
            handler = handler.__name__

        futures = []
        kwargs.update(dict(_tell=True, _wait=False))
        for ep in self.get_schedulers():
            ref = self.ctx.actor_ref(SessionManagerActor.default_uid(), address=ep)
            futures.append(ref.broadcast_sessions(handler, *args, **kwargs))
        [f.result() for f in futures]

    def _broadcast_workers(self, handler, *args, **kwargs):
        from ..worker.execution import ExecutionActor

        if hasattr(handler, '__name__'):
            handler = handler.__name__

        kwargs.update(dict(_tell=True, _wait=False))
        for w in self._meta_cache.keys():
            ref = self.ctx.actor_ref(ExecutionActor.default_uid(), address=w)
            getattr(ref, handler)(*args, **kwargs)

    def allocate_resource(self, session_id, op_key, endpoint, alloc_dict, log_fail=False):
        """
        Try allocate resource for operands
        :param session_id: session id
        :param op_key: operand key
        :param endpoint: worker endpoint
        :param alloc_dict: allocation dict, listing resources needed by the operand
        :param log_fail: log assign failure
        :return: True if allocated successfully
        """
        worker_stats = self._meta_cache[endpoint]['hardware']
        if endpoint not in self._worker_allocations:
            self._worker_allocations[endpoint] = dict()
        if session_id not in self._worker_allocations[endpoint]:
            self._worker_allocations[endpoint][session_id] = dict()
        worker_allocs = self._worker_allocations[endpoint][session_id]

        res_used = defaultdict(lambda: 0)
        free_alloc_keys = []
        for alloc in worker_allocs.values():
            for k, v in alloc[0].items():
                res_used[k] += v
        for k in free_alloc_keys:
            self.deallocate_resource(session_id, k, endpoint)

        for k, v in alloc_dict.items():
            res_used[k] += v

        sufficient = True
        for k, v in res_used.items():
            if res_used[k] > worker_stats.get(k, 0):
                sufficient = False
                if log_fail:
                    logger.warning('Failed to assign %s resource for %s, res_used: %s, worker stat: %s',
                                   k, op_key, res_used[k], worker_stats.get(k, 0))
                break

        if sufficient:
            self._worker_allocations[endpoint][session_id][op_key] = (alloc_dict, time.time())
            return True
        else:
            return False

    def deallocate_resource(self, session_id, op_key, endpoint):
        """
        Deallocate resource
        :param session_id: session id
        :param op_key: operand key
        :param endpoint: worker endpoint
        """
        try:
            del self._worker_allocations[endpoint][session_id][op_key]
        except KeyError:
            pass

    def get_allocations(self):
        return self._worker_allocations



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .utils import SchedulerActor


class CustomLogMetaActor(SchedulerActor):
    def __init__(self):
        super().__init__()

        self._custom_log_path_store = dict()

    def record_custom_log_path(self, session_id: str, tileable_op_key: str,
                               chunk_op_key: str, worker_address: str,
                               log_path: str):
        key = (session_id, tileable_op_key)
        if key not in self._custom_log_path_store:
            self._custom_log_path_store[key] = dict()
        self._custom_log_path_store[key][chunk_op_key] = (worker_address, log_path)

    def get_tileable_op_log_paths(self, session_id: str, tileable_op_key: str):
        return self._custom_log_path_store.get((session_id, tileable_op_key))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...tensor import arithmetic
from ...tensor import reduction
from ...tensor.fuse import TensorNeFuseChunk
from ...tensor.fuse.ne import NUMEXPR_INSTALLED
from ...utils import build_fuse_chunk

REDUCTION_OP = {reduction.TensorSum, reduction.TensorProd,
                reduction.TensorMax, reduction.TensorMin}
SUPPORT_OP = {
    arithmetic.TensorAdd,
    arithmetic.TensorSubtract,
    arithmetic.TensorMultiply,
    arithmetic.TensorDivide,
    arithmetic.TensorPower,
    arithmetic.TensorMod,
    arithmetic.TensorNegative,
    arithmetic.TensorAbs,
    arithmetic.TensorConj,
    arithmetic.TensorExp,
    arithmetic.TensorLog,
    arithmetic.TensorLog10,
    arithmetic.TensorExpm1,
    arithmetic.TensorLog1p,
    arithmetic.TensorSqrt,

    arithmetic.TensorEqual,
    arithmetic.TensorNotEqual,
    arithmetic.TensorLessThan,
    arithmetic.TensorLessEqual,
    arithmetic.TensorGreaterThan,
    arithmetic.TensorGreaterEqual,

    arithmetic.TensorSin,
    arithmetic.TensorCos,
    arithmetic.TensorTan,
    arithmetic.TensorArcsin,
    arithmetic.TensorArccos,
    arithmetic.TensorArctan,
    arithmetic.TensorSinh,
    arithmetic.TensorCosh,
    arithmetic.TensorTanh,
    arithmetic.TensorArcsinh,
    arithmetic.TensorArccosh,
    arithmetic.TensorArctanh,

    arithmetic.TensorLshift,
    arithmetic.TensorRshift,

    arithmetic.TensorTreeAdd,
    arithmetic.TensorTreeMultiply,

    reduction.TensorSum,
    reduction.TensorProd,
    reduction.TensorMax,
    reduction.TensorMin
}


def _check_reduction_axis(node):
    return len(node.op.axis) == 1 or len(node.op.axis) == node.ndim


def _support(node):
    op_type = type(node.op)
    if op_type in REDUCTION_OP:
        return _check_reduction_axis(node)
    return op_type in SUPPORT_OP


def _transfer_op(node):
    op = node.op
    if type(op) in REDUCTION_OP and not _check_reduction_axis(node):
        return op
    return op


class NeRuntimeOptimizer:
    def __init__(self, graph):
        self._graph = graph

    @classmethod
    def is_available(cls):
        return NUMEXPR_INSTALLED

    def optimize(self, keys=None):
        self.compose(keys=keys)

    def _compose_graph(self, composes):
        graph = self._graph
        composed_nodes = []

        for c in composes:
            head_node = c[0]
            tail_node = c[-1]

            composed_chunk = build_fuse_chunk(
                c, TensorNeFuseChunk,
                op_kw={'dtype': tail_node.dtype}).data
            graph.add_node(composed_chunk)
            for node in graph.iter_successors(tail_node):
                graph.add_edge(composed_chunk, node)
            for node in graph.iter_predecessors(head_node):
                graph.add_edge(node, composed_chunk)
            for node in c:
                graph.remove_node(node)
            composed_nodes.append(composed_chunk)

        return composed_nodes

    def compose(self, keys=None):
        composes = []
        explored = set()
        keys = set(keys or [])

        graph = self._graph
        for v in graph.topological_iter():
            if v.op.gpu or v.op.sparse:
                # break out
                return []
            if type(v.op) not in SUPPORT_OP or v.key in keys:
                continue
            if v in explored or type(v.op) in REDUCTION_OP:  # TODO: check logic here
                continue
            if graph.count_successors(v) != 1:
                continue
            selected = [v]
            # add successors
            cur_node = graph.successors(v)[0]
            while graph.count_predecessors(cur_node) == 1 \
                    and _support(cur_node) and cur_node.key not in keys:
                selected.append(cur_node)
                if graph.count_successors(cur_node) != 1 \
                        or type(cur_node.op) in REDUCTION_OP:
                    break
                else:
                    cur_node = graph.successors(cur_node)[0]
            if len(selected) > 1:
                explored.update(selected)
                composes.append(list(selected))
        return self._compose_graph(composes)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes
from ...context import get_context
from ...serialize import KeyField, Float64Field, Int64Field, BoolField
from ...tensor.datasource import tensor as astensor
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape, ceildiv
from ..operands import TensorOperandMixin, TensorOperand


class RebalanceMixin:
    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def __call__(self, df_or_series):
        self._output_types = df_or_series.op.output_types
        return self.new_tileable([df_or_series], kws=[df_or_series.params])

    def _get_input_object(self):
        raise NotImplementedError

    @classmethod
    def tile(cls, op: "RebalanceMixin"):
        in_obj = op._get_input_object()
        ctx = get_context()

        if ctx is None and op.factor is not None:
            return [in_obj]

        check_chunks_unknown_shape([in_obj], TilesError)

        size = in_obj.shape[op.axis]
        if op.factor is not None:
            cluster_cpu_count = ctx.get_total_ncores()
            assert cluster_cpu_count > 0
            expect_n_chunk = int(cluster_cpu_count * op.factor)
        else:
            expect_n_chunk = op.num_partitions

        expect_chunk_size = max(ceildiv(size, expect_n_chunk), 1)
        return [in_obj.rechunk(
            {op.axis: expect_chunk_size}, reassign_worker=op.reassign_worker)._inplace_tile()]


class TensorRebalance(RebalanceMixin, TensorOperandMixin, TensorOperand):
    _op_type_ = opcodes.REBALANCE

    _input = KeyField('input')
    _factor = Float64Field('factor')
    _axis = Int64Field('axis')
    _num_partitions = Int64Field('num_partitions')
    _reassign_worker = BoolField('reassign_worker')

    def __init__(self, input=None, factor=None, axis=None,  # pylint: disable=redefined-builtin
                 num_partitions=None, reassign_worker=None, output_types=None, **kw):
        super().__init__(_input=input, _factor=factor, _axis=axis, _num_partitions=num_partitions,
                         _output_types=output_types, _reassign_worker=reassign_worker, **kw)

    @property
    def input(self):
        return self._input

    @property
    def factor(self):
        return self._factor

    @property
    def axis(self):
        return self._axis

    @property
    def num_partitions(self):
        return self._num_partitions

    @property
    def reassign_worker(self):
        return self._reassign_worker

    def _get_input_object(self):
        return astensor(self.inputs[0])


def rebalance(tensor, factor=None, axis=0, num_partitions=None, reassign_worker=True):
    """
    Make Data more balanced across entire cluster.

    Parameters
    ----------
    factor : float
        Specified so that number of chunks after balance is
        total CPU count of cluster * factor.
    axis : int
        The axis to rebalance.
    num_partitions : int
        Specified so the number of chunks are at most
        num_partitions.
    reassign_worker : bool
        If True, workers will be reassigned.

    Returns
    -------
    Series or DataFrame
        Result of DataFrame or Series after rebalanced.
    """
    if num_partitions is None:
        factor = factor if factor is not None else 1.2

    op = TensorRebalance(input=tensor, factor=factor, axis=axis, num_partitions=num_partitions,
                         reassign_worker=reassign_worker)
    return op(tensor)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys

from . import kvstore
from .actors import FunctionActor
from .promise import PromiseActor
from .lib.uhashring import HashRing
from .utils import to_str


SCHEDULER_PATH = '/schedulers'
INITIAL_SCHEDULER_FILE = '/tmp/mars-initial-schedulers.tmp'
logger = logging.getLogger(__name__)


def create_hash_ring(schedulers):
    return HashRing(nodes=schedulers, hash_fn='ketama')


def get_scheduler(hash_ring, key, size=1):
    if hash_ring is None:
        return None
    if size == 1:
        return hash_ring.get_node(key)
    return tuple(it['nodename'] for it in hash_ring.range(key, size=size))


class StaticClusterDiscoverer(object):
    dynamic = False

    def __init__(self, schedulers):
        self._schedulers = schedulers

    def __reduce__(self):
        return type(self), (self._schedulers,)

    def get_schedulers(self):
        return self._schedulers

    def watch_schedulers(self):
        raise NotImplementedError


class KVStoreClusterDiscoverer(object):
    dynamic = True

    def __init__(self, address):
        self._address = address
        self._client = kvstore.get(address)
        if isinstance(self._client, kvstore.LocalKVStore):
            raise ValueError(f'etcd_addr should not be a local address, got {address}')

    def __reduce__(self):
        return type(self), (self._address,)

    @staticmethod
    def _resolve_schedulers(result):
        return [to_str(s.key.rsplit('/', 1)[1]) for s in result.children]

    def get_schedulers(self):
        try:
            return self._resolve_schedulers(self._client.read(SCHEDULER_PATH))
        except KeyError:
            return next(self.watch_schedulers())

    def watch_schedulers(self):
        for result in self._client.eternal_watch(SCHEDULER_PATH):
            yield self._resolve_schedulers(result)


class _ClusterInfoWatchActor(FunctionActor):
    def __init__(self, discoverer, cluster_info_ref):
        self._discoverer = discoverer
        self._cluster_info_ref = cluster_info_ref

    def post_create(self):
        self._cluster_info_ref = self.ctx.actor_ref(self._cluster_info_ref)

    def get_schedulers(self):
        return self._discoverer.get_schedulers()

    def watch_schedulers(self):
        for schedulers in self._discoverer.watch_schedulers():
            self._cluster_info_ref.set_schedulers(schedulers, _tell=True)


class ClusterInfoActor(FunctionActor):
    def __init__(self, discoverer, distributed=True):
        if isinstance(discoverer, list):
            discoverer = StaticClusterDiscoverer(discoverer)

        self._discoverer = discoverer
        self._distributed = distributed
        self._hash_ring = None
        self._watcher = None
        self._schedulers = []
        self._observer_refs = dict()

    @classmethod
    def default_uid(cls):
        raise NotImplementedError

    def post_create(self):
        logger.debug('Actor %s running in process %d', self.uid, os.getpid())

        if self._discoverer.dynamic:
            watcher = self._watcher = self.ctx.create_actor(
                _ClusterInfoWatchActor, self._discoverer, self.ref())
            watcher.watch_schedulers(_tell=True)
        self._schedulers = self._discoverer.get_schedulers()

        self._hash_ring = create_hash_ring(self._schedulers)

    def pre_destroy(self):
        if self._watcher:
            self.ctx.destroy_actor(self._watcher)

    def register_observer(self, observer, fun_name):
        self._observer_refs[(observer.uid, observer.address)] = (self.ctx.actor_ref(observer), fun_name)

    def unregister_observer(self, observer):
        self._observer_refs.pop((observer.uid, observer.address), None)

    def get_schedulers(self):
        return self._schedulers

    def set_schedulers(self, schedulers):
        logger.debug('Setting schedulers %r', schedulers)
        self._schedulers = schedulers
        self._hash_ring = create_hash_ring(self._schedulers)

        for observer_ref, fun_name in self._observer_refs.values():
            # notify the observers to update the new scheduler list
            getattr(observer_ref, fun_name)(schedulers, _tell=True, _wait=False)

    def get_scheduler(self, key, size=1):
        if len(self._schedulers) == 1 and size == 1:
            return self._schedulers[0]
        return get_scheduler(self._hash_ring, key, size=size)

    def is_distributed(self):
        return self._distributed

    def rescale_workers(self, new_scale):
        try:
            return self._discoverer.rescale_workers(new_scale)
        except AttributeError:
            raise NotImplementedError from None


class HasClusterInfoActor(PromiseActor):
    cluster_info_uid = None

    def __init__(self):
        super().__init__()

        # the scheduler list
        if sys.platform.startswith('linux') and os.path.exists(INITIAL_SCHEDULER_FILE):
            # there is predefined scheduler list, read first
            with open(INITIAL_SCHEDULER_FILE, 'r') as scheduler_file:
                self._schedulers = scheduler_file.read().strip().split(',')
        else:
            self._schedulers = None
        self._hash_ring = None

        self._cluster_info_ref = None

    def get_schedulers(self):
        return self._schedulers

    def set_schedulers(self, schedulers):
        self._schedulers = schedulers
        if len(schedulers) > 1:
            self._hash_ring = create_hash_ring(schedulers)

    def set_cluster_info_ref(self, set_schedulers_fun_name=None):
        set_schedulers_fun_name = set_schedulers_fun_name or self.set_schedulers.__name__

        # cluster_info_actor is created when scheduler initialized
        self._cluster_info_ref = self.ctx.actor_ref(self.cluster_info_uid)
        # when some schedulers lost, notification will be received
        self._cluster_info_ref.register_observer(
            self.ref(), set_schedulers_fun_name, _tell=True, _wait=False)
        if not self._schedulers:
            self.set_schedulers(self._cluster_info_ref.get_schedulers())

    def unset_cluster_info_ref(self):
        self._cluster_info_ref = self.ctx.actor_ref(self.cluster_info_uid)
        # when some schedulers lost, notification will be received
        self._cluster_info_ref.unregister_observer(self.ref(), _tell=True, _wait=False)

    def get_actor_ref(self, key):
        addr = self.get_scheduler(key)
        return self.ctx.actor_ref(addr, key)

    def get_promise_ref(self, key):
        addr = self.get_scheduler(key)
        return self.promise_ref(addr, key)

    def get_scheduler(self, key, size=1):
        if len(self._schedulers) == 1:
            return self._schedulers[0]

        return get_scheduler(self._hash_ring, key, size=size)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from .... import opcodes as OperandDef
from ....core import ExecutableTuple, get_output_types
from ....serialize.core import KeyField, Float64Field, ListField, BoolField
from ....tensor.core import TENSOR_TYPE, TENSOR_CHUNK_TYPE
from ....tensor import tensor as astensor
from ....dataframe.core import DATAFRAME_TYPE
from ...operands import LearnOperand, LearnOperandMixin
from ...utils import convert_to_tensor_or_dataframe, concat_chunks


class ToDMatrix(LearnOperand, LearnOperandMixin):
    _op_type_ = OperandDef.TO_DMATRIX

    _data = KeyField('data')
    _label = KeyField('label')
    _missing = Float64Field('missing')
    _weight = KeyField('weight')
    _feature_names = ListField('feature_names')
    _feature_types = ListField('feature_types')
    _multi_output = BoolField('multi_output')

    def __init__(self, data=None, label=None, missing=None, weight=None, feature_names=None,
                 feature_types=None, multi_output=None, gpu=None, output_types=None, **kw):
        super().__init__(_data=data, _label=label, _missing=missing, _weight=weight,
                         _feature_names=feature_names, _feature_types=feature_types,
                         _gpu=gpu, _multi_output=multi_output, _output_types=output_types,
                         **kw)

    @property
    def output_limit(self):
        if self._multi_output:
            return 1 + (self._label is not None) + (self._weight is not None)
        return 1

    @property
    def data(self):
        return self._data

    @property
    def label(self):
        return self._label

    @property
    def missing(self):
        return self._missing

    @property
    def weight(self):
        return self._weight

    @property
    def feature_names(self):
        return self._feature_names

    @property
    def feature_types(self):
        return self._feature_types

    @property
    def multi_output(self):
        return self._multi_output

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._data = self._inputs[0]
        has_label = self._label is not None
        if has_label:
            self._label = self._inputs[1]
        if self._weight is not None:
            i = 1 if not has_label else 2
            self._weight = self._inputs[i]

    @staticmethod
    def _get_kw(obj):
        if isinstance(obj, TENSOR_TYPE + TENSOR_CHUNK_TYPE):
            return {'shape': obj.shape,
                    'dtype': obj.dtype,
                    'order': obj.order}
        else:
            return {'shape': obj.shape,
                    'dtypes': obj.dtypes,
                    'index_value': obj.index_value,
                    'columns_value': obj.columns_value}

    def __call__(self):
        inputs = [self._data]
        kws = []
        kw = self._get_kw(self._data)
        if not self._multi_output:
            kw['type'] = 'data'
        kws.append(kw)
        if self._label is not None:
            inputs.append(self._label)
            if self._multi_output:
                kw = self._get_kw(self._label)
                kw['type'] = 'label'
                kws.append(kw)
        if self._weight is not None:
            inputs.append(self._weight)
            if self._multi_output:
                kw = self._get_kw(self._weight)
                kw['type'] = 'weight'
                kws.append(kw)
        if not self.output_types:
            self.output_types = get_output_types(*inputs)

        tileables = self.new_tileables(inputs, kws=kws)
        if not self._multi_output:
            return tileables[0]
        return tileables

    @classmethod
    def _tile_multi_output(cls, op):
        data, label, weight = op.data, op.label, op.weight

        if data.chunk_shape[1] > 1:
            # make sure data's second dimension has only 1 chunk
            data = data.rechunk({1: data.shape[1]})._inplace_tile()

        nsplit = data.nsplits[0]
        # rechunk label
        if label is not None:
            label = label.rechunk({0: nsplit})._inplace_tile()
        # rechunk weight
        if weight is not None:
            weight = weight.rechunk({0: nsplit})._inplace_tile()

        out_chunkss = [[] for _ in range(op.output_limit)]
        for i in range(len(nsplit)):
            data_chunk = data.cix[i, 0]
            inps = [data_chunk]
            kws = []
            chunk_op = op.copy().reset_key()
            chunk_op._data = data_chunk
            data_kw = cls._get_kw(data_chunk)
            data_kw['index'] = data_chunk.index
            kws.append(data_kw)
            if label is not None:
                label_chunk = chunk_op._label = label.cix[i, ]
                inps.append(label_chunk)
                kw = cls._get_kw(label_chunk)
                kw['index'] = label_chunk.index
                kw['type'] = 'label'
                kws.append(kw)
            if weight is not None:
                weight_chunk = chunk_op._weight = weight.cix[i, ]
                inps.append(weight_chunk)
                kw = cls._get_kw(weight_chunk)
                kw['index'] = weight_chunk.index
                kw['type'] = 'weight'
                kws.append(kw)
            out_chunks = chunk_op.new_chunks(inps, kws=kws)
            for i, out_chunk in enumerate(out_chunks):
                out_chunkss[i].append(out_chunk)

        new_op = op.copy()
        params = [out.params.copy() for out in op.outputs]
        types = ['data', 'label', 'weight']
        for i, inp in enumerate([data, label, weight]):
            if inp is None:
                continue
            params[i]['nsplits'] = inp.nsplits
            params[i]['chunks'] = out_chunkss[i]
            params[i]['type'] = types[i]
        return new_op.new_tileables(op.inputs, kws=params)

    @classmethod
    def _tile_single_output(cls, op):
        from ....context import get_context, RunningMode

        data, label, weight = op.data, op.label, op.weight

        ctx = get_context()
        if ctx.running_mode != RunningMode.distributed:
            # for local and local cluster, just concat all data into one
            data_concat_chunk = data.op.concat_tileable_chunks(data).chunks[0]
            inps = [data_concat_chunk]
            label_concat_chunk = None
            if label is not None:
                label_concat_chunk = label.op.concat_tileable_chunks(label).chunks[0]
                inps.append(label_concat_chunk)
            weight_concat_chunk = None
            if weight is not None:
                weight_concat_chunk = weight.op.concat_tileable_chunks(weight).chunks[0]
                inps.append(weight_concat_chunk)
            chunk_op = ToDMatrix(data=data_concat_chunk, label=label_concat_chunk,
                                 missing=op.missing, weight=weight_concat_chunk,
                                 feature_names=op.feature_names,
                                 feature_types=op.feature_types,
                                 multi_output=False, output_types=op.output_types)
            out_chunks = [chunk_op.new_chunk(inps, **data_concat_chunk.params)]
            nsplits = tuple((s,) for s in data_concat_chunk.shape)
        else:
            # for distributed, we should concat the chunks
            # which allocated on the same worker into one
            data_chunk_metas = ctx.get_chunk_metas([c.key for c in data.chunks])
            data_chunk_workers = [m.workers[0] for m in data_chunk_metas]
            worker_to_chunks = dict()
            for i, worker in enumerate(data_chunk_workers):
                size = 1 + (label is not None) + (weight is not None)
                if worker not in worker_to_chunks:
                    worker_to_chunks[worker] = [[] for _ in range(size)]
                worker_to_chunks[worker][0].append(data.chunks[i])
                if label is not None:
                    worker_to_chunks[worker][1].append(label.chunks[i])
                if weight is not None:
                    worker_to_chunks[worker][-1].append(weight.chunks[i])
            ind = itertools.count(0)
            out_chunks = []
            for worker, chunks in worker_to_chunks.items():
                data_chunk = concat_chunks(chunks[0])
                inps = [data_chunk]
                label_chunk = None
                if label is not None:
                    label_chunk = concat_chunks(chunks[1])
                    inps.append(label_chunk)
                weight_chunk = None
                if weight is not None:
                    weight_chunk = concat_chunks(chunks[2])
                    inps.append(weight_chunk)
                chunk_op = ToDMatrix(data=data_chunk, label=label_chunk, missing=op.missing,
                                     weight=weight_chunk, feature_names=op.feature_names,
                                     feature_types=op.feature_types, multi_output=False,
                                     output_types=op.output_types)
                kws = data_chunk.params
                kws['index'] = (next(ind), 0)
                out_chunks.append(chunk_op.new_chunk(inps, **kws))
            nsplits = (tuple(c.shape[0] for c in out_chunks), (out_chunks[0].shape[1],))

        new_op = op.copy()
        kw = op.outputs[0].params
        kw['chunks'] = out_chunks
        kw['nsplits'] = nsplits
        return new_op.new_tileables(op.inputs, kws=[kw])

    @classmethod
    def tile(cls, op):
        if op.multi_output:
            return cls._tile_multi_output(op)
        else:
            return cls._tile_single_output(op)

    @staticmethod
    def get_xgb_dmatrix(tup):
        from xgboost import DMatrix

        data, label, weight, missing, feature_names, feature_types = tup
        data = data.spmatrix if hasattr(data, 'spmatrix') else data
        return DMatrix(data, label=label, missing=missing, weight=weight,
                       feature_names=feature_names, feature_types=feature_types,
                       nthread=-1)

    @staticmethod
    def _from_ctx_if_not_none(ctx, chunk):
        if chunk is None:
            return chunk
        return ctx[chunk.key]

    def execute(cls, ctx, op):
        if op.multi_output:
            outs = op.outputs
            ctx[outs[0].key] = ctx[op.inputs[0].key]
            if op.label is not None:
                ctx[outs[1].key] = ctx[op.inputs[1].key]
            if op.weight is not None:
                ctx[outs[-1].key] = ctx[op.inputs[-1].key]
            return
        else:
            ctx[op.outputs[0].key] = (
                cls._from_ctx_if_not_none(ctx, op.data),
                cls._from_ctx_if_not_none(ctx, op.label),
                cls._from_ctx_if_not_none(ctx, op.weight),
                op.missing,
                op.feature_names,
                op.feature_types
            )


def check_data(data):
    data = convert_to_tensor_or_dataframe(data)
    if data.ndim != 2:
        raise ValueError(f'Expecting 2-d data, got: {data.ndim}-d')

    return data


def to_dmatrix(data, label=None, missing=None, weight=None,
               feature_names=None, feature_types=None, session=None, run_kwargs=None):
    data = check_data(data)
    if label is not None:
        label = convert_to_tensor_or_dataframe(label)
        if isinstance(label, DATAFRAME_TYPE):
            label = label.iloc[:, 0]
        label = astensor(label)
        if label.ndim != 1:
            raise ValueError(f'Expecting 1-d label, got: {label.ndim}-d')
    if weight is not None:
        weight = convert_to_tensor_or_dataframe(weight)
        if isinstance(weight, DATAFRAME_TYPE):
            weight = weight.iloc[:, 0]
        weight = astensor(weight)
        if weight.ndim != 1:
            raise ValueError(f'Expecting 1-d weight, got {weight.ndim}-d')

    op = ToDMatrix(data=data, label=label, missing=missing, weight=weight,
                   feature_names=feature_names, feature_types=feature_types,
                   gpu=data.op.gpu, multi_output=True,
                   output_types=get_output_types(data, label, weight))
    outs = ExecutableTuple(op())
    # Execute first, to make sure the counterpart chunks of data, label and weight are co-allocated
    outs.execute(session=session, **(run_kwargs or dict()))

    data = outs[0]
    label = None if op.label is None else outs[1]
    weight = None if op.weight is None else outs[-1]
    # If not multiple outputs, try to collect the chunks on same worker into one
    # to feed the data into XGBoost for training.
    op = ToDMatrix(data=data, label=label, missing=missing, weight=weight,
                   feature_names=feature_names, feature_types=feature_types,
                   gpu=data.op.gpu, multi_output=False,
                   output_types=get_output_types(data))
    return op()


MarsDMatrix = to_dmatrix



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..operands import Operand, TileableOperandMixin, Fuse, FuseChunkMixin, \
    MapReduceOperand, ShuffleProxy, OutputType
from ..tensor.core import TENSOR_TYPE, TENSOR_CHUNK_TYPE
from ..tensor.operands import TensorOperandMixin
from ..tensor.fuse import TensorFuseChunk
from ..dataframe.core import TILEABLE_TYPE as DATAFRAME_TYPE, \
    CHUNK_TYPE as DATAFRAME_CHUNK_TYPE
from ..dataframe.operands import DataFrameOperandMixin, DataFrameFuseChunk


LearnOperand = Operand


class LearnOperandMixin(TileableOperandMixin):
    __slots__ = ()
    _op_module_ = 'learn'

    @classmethod
    def concat_tileable_chunks(cls, tileable):
        if isinstance(tileable, TENSOR_TYPE):
            return TensorOperandMixin.concat_tileable_chunks(tileable)
        elif isinstance(tileable, DATAFRAME_TYPE):
            return DataFrameOperandMixin.concat_tileable_chunks(tileable)
        else:
            # op has to implement its logic of `concat_tileable_chunks`
            raise NotImplementedError

    @classmethod
    def create_tileable_from_chunks(cls, chunks, inputs=None, **kw):
        if isinstance(chunks[0], TENSOR_CHUNK_TYPE):
            return TensorOperandMixin.create_tileable_from_chunks(
                chunks, inputs=inputs, **kw)
        elif isinstance(chunks[0], DATAFRAME_CHUNK_TYPE):
            return DataFrameOperandMixin.create_tileable_from_chunks(
                chunks, inputs=inputs, **kw)
        else:
            # op has to implement its logic of `create_tileable_from_chunks`
            raise NotImplementedError

    def get_fuse_op_cls(self, obj):
        if isinstance(obj, (TENSOR_TYPE, TENSOR_CHUNK_TYPE)):
            return TensorFuseChunk
        elif isinstance(obj, (DATAFRAME_TYPE, DATAFRAME_CHUNK_TYPE)):
            return DataFrameFuseChunk
        else:
            return LearnObjectFuseChunk


class LearnObjectFuseChunkMixin(FuseChunkMixin, LearnOperandMixin):
    __slots__ = ()

    _output_type_ = OutputType.object


class LearnObjectFuseChunk(LearnObjectFuseChunkMixin, Fuse):
    pass


class LearnShuffleProxy(ShuffleProxy, LearnOperandMixin):
    def __init__(self, output_types=None, **kw):
        super().__init__(_output_types=output_types, **kw)
        if not self.output_types:
            self.output_types = [OutputType.object]

    @classmethod
    def execute(cls, ctx, op):
        pass


LearnMapReduceOperand = MapReduceOperand



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pickle import dumps, loads

import numpy as np
cimport numpy as np
cimport cython
from libc.string cimport memcpy
from cpython.bytearray cimport PyByteArray_AS_STRING, PyByteArray_GET_SIZE, PyByteArray_Resize

from ..core cimport ActorRef
from ...lib.tblib import pickling_support

pickling_support.install()

# Internal message types includes:
# 1) create actor
# 2) destroy actor
# 3) has actor
# 4) result that return back by the actor in a different process or machine
# 5) error that throws back by the actor in a different process or machine
# 6) send all, the send actor message
# 7) tell all, the tell actor message
cpdef BYTE_t CREATE_ACTOR = MessageType.create_actor
cpdef BYTE_t DESTROY_ACTOR = MessageType.destroy_actor
cpdef BYTE_t HAS_ACTOR = MessageType.has_actor
cpdef BYTE_t RESULT = MessageType.result
cpdef BYTE_t ERROR = MessageType.error
cpdef BYTE_t SEND_ALL = MessageType.send_all
cpdef BYTE_t TELL_ALL = MessageType.tell_all


cpdef enum MessageSerialType:
    # Internal message serialized type
    null = 0
    raw_bytes = 1
    pickle = 2


cdef BYTE_t NONE = MessageSerialType.null
cdef BYTE_t RAW_BYTES = MessageSerialType.raw_bytes
cdef BYTE_t PICKLE = MessageSerialType.pickle


cdef INT32_t DEFAULT_PROTOCOL = 0


cdef class _BASE_ACTOR_MESSAGE:
    cdef public INT32_t message_type
    cdef public bytes message_id
    cdef public INT32_t from_index
    cdef public INT32_t to_index


cdef class CREATE_ACTOR_MESSAGE(_BASE_ACTOR_MESSAGE):
    cdef public ActorRef actor_ref
    cdef public object actor_cls
    cdef public tuple args
    cdef public dict kwargs

    def __init__(self, INT32_t message_type=-1, bytes message_id=None,
                 INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
                 object actor_cls=None, tuple args=None, dict kwargs=None):
        self.message_type = message_type
        self.message_id = message_id
        self.from_index = from_index
        self.to_index = to_index
        self.actor_ref = actor_ref
        self.actor_cls = actor_cls
        self.args = args
        self.kwargs = kwargs


cdef class DESTROY_ACTOR_MESSAGE(_BASE_ACTOR_MESSAGE):
    cdef public ActorRef actor_ref

    def __init__(self, INT32_t message_type=-1, bytes message_id=None,
                 INT32_t from_index=0, INT32_t to_index=0, object actor_ref=None):
        self.message_type = message_type
        self.message_id = message_id
        self.from_index = from_index
        self.to_index = to_index
        self.actor_ref = actor_ref


cdef class HAS_ACTOR_MESSAGE(_BASE_ACTOR_MESSAGE):
    cdef public ActorRef actor_ref

    def __init__(self, INT32_t message_type=-1, bytes message_id=None,
                 INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None):
        self.message_type = message_type
        self.message_id = message_id
        self.from_index = from_index
        self.to_index = to_index
        self.actor_ref = actor_ref


cdef class RESULT_MESSAGE(_BASE_ACTOR_MESSAGE):
    cdef public object result

    def __init__(self, INT32_t message_type=-1, bytes message_id=None,
                 INT32_t from_index=0, INT32_t to_index=0, object result=None):
        self.message_type = message_type
        self.message_id = message_id
        self.from_index = from_index
        self.to_index = to_index
        self.result = result


cdef class ERROR_MESSAGE(_BASE_ACTOR_MESSAGE):
    cdef public object error_type
    cdef public object error
    cdef public object traceback

    def __init__(self, INT32_t message_type=-1, bytes message_id=None,
                 INT32_t from_index=0, INT32_t to_index=0, object error_type=None,
                 object error=None, object traceback=None):
        self.message_type = message_type
        self.message_id = message_id
        self.from_index = from_index
        self.to_index = to_index
        self.error_type = error_type
        self.error = error
        self.traceback = traceback


cdef class SEND_MESSAGE(_BASE_ACTOR_MESSAGE):
    cdef public ActorRef actor_ref
    cdef public object message

    def __init__(self, object message_type=-1, bytes message_id=None,
                 INT32_t from_index=0, INT32_t to_index=0, ActorRef actor_ref=None,
                 object message=None):
        self.message_type = message_type
        self.message_id = message_id
        self.from_index = from_index
        self.to_index = to_index
        self.actor_ref = actor_ref
        self.message = message


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void _pack_byte(BYTE_t val, bytearray arr):
    cdef char *ptr
    cdef Py_ssize_t size = PyByteArray_GET_SIZE(arr)
    PyByteArray_Resize(arr, size + 1)
    ptr = PyByteArray_AS_STRING(arr) + size
    ptr[0] = val


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline BYTE_t _unpack_byte(const char *s) nogil:
    return s[0]


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void _pack_short(INT16_t val, bytearray arr):
    cdef char *ptr
    cdef Py_ssize_t size = PyByteArray_GET_SIZE(arr)
    PyByteArray_Resize(arr, size + sizeof(INT16_t))
    ptr = PyByteArray_AS_STRING(arr) + size
    memcpy(ptr, &val, sizeof(INT16_t))


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline INT16_t _unpack_short(const char *s) nogil:
    cdef INT16_t ret
    memcpy(&ret, s, sizeof(INT16_t))
    return ret


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void _pack_int(INT32_t val, bytearray arr):
    cdef char *ptr
    cdef Py_ssize_t size = PyByteArray_GET_SIZE(arr)
    PyByteArray_Resize(arr, size + sizeof(INT32_t))
    ptr = PyByteArray_AS_STRING(arr) + size
    memcpy(ptr, &val, sizeof(INT32_t))


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline INT32_t _unpack_int(const char *s) nogil:
    cdef INT32_t ret
    memcpy(&ret, s, sizeof(INT32_t))
    return ret


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void _pack_long(INT64_t val, bytearray arr):
    cdef char *ptr
    cdef Py_ssize_t size = PyByteArray_GET_SIZE(arr)
    PyByteArray_Resize(arr, size + sizeof(INT64_t))
    ptr = PyByteArray_AS_STRING(arr) + size
    memcpy(ptr, &val, sizeof(INT64_t))


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline INT64_t _unpack_long(const char *s) nogil:
    cdef INT64_t ret
    memcpy(&ret, s, sizeof(INT64_t))
    return ret


cdef object np_random_bytes = np.random.bytes


cdef inline bytes new_message_id():
    return np_random_bytes(32)


cdef inline void _pack_message_type(INT32_t message_type, bytearray buf, INT32_t protocol=DEFAULT_PROTOCOL):
    cdef INT32_t value

    # use 1 byte to represent protocol and message type, from left to right, 0-2, protocol, 3-7 message_type
    value = (protocol << 5) | message_type
    _pack_byte(value, buf)


cdef inline INT32_t _unpack_message_type_value(const char *binary, size_t* pos) nogil:
    cdef INT32_t value
    cdef INT32_t protocol

    value = binary[pos[0]]
    pos[0] += 1
    protocol = value >> 5
    if protocol != 0:
        raise NotImplementedError('Unsupported protocol')

    return value & ((1 << 5) - 1)


cdef inline object _unpack_message_type(const char *binary, size_t* pos):
    return MessageType(_unpack_message_type_value(binary, pos))


cpdef INT32_t unpack_message_type_value(bytes binary):
    cdef size_t pos = 0
    return _unpack_message_type_value(binary, &pos)


cpdef object unpack_message_type(bytes binary):
    return MessageType(unpack_message_type_value(binary))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef inline void _skip_message_type(const char *binary, size_t* pos) nogil:
    pos[0] += 1


cpdef bytes unpack_message_id(bytes binary):
    cdef size_t pos = 1

    return _unpack_message_id(binary, &pos)


cdef inline void _pack_message_id(bytes message_id, bytearray buf):
    _pack_object(message_id, buf)


cdef inline bytes _unpack_message_id(const char *buf, size_t* pos):
    return _unpack_object(buf, pos)


cdef inline void _skip_message_id(const char *buf, size_t* pos) nogil:
    _skip_object(buf, pos)


cdef inline void _pack_object(object obj, bytearray buf) except *:
    cdef BYTE_t st
    cdef bytes m

    if obj is None:
        _pack_byte(NONE, buf)
        return

    if isinstance(obj, bytes):
        st = RAW_BYTES
        m = obj
    else:
        st = PICKLE
        m = dumps(obj)

    _pack_byte(st, buf)
    _pack_long(len(m), buf)
    buf.extend(m)


cdef inline object _unpack_object(const char *binary, size_t* pos):
    cdef BYTE_t st
    cdef size_t size
    cdef bytes ret_bytes

    st = binary[pos[0]]
    pos[0] += 1

    if st == NONE:
        return None

    size = _unpack_long(binary + pos[0])
    pos[0] += 8

    ret_bytes = binary[pos[0]:pos[0] + size]
    pos[0] += size
    if st == RAW_BYTES:
        return ret_bytes
    else:
        return loads(ret_bytes)


cdef inline void _skip_object(const char *binary, size_t* pos) nogil:
    cdef BYTE_t st
    cdef size_t size

    st = binary[pos[0]]
    pos[0] += 1

    if st == NONE:
        return

    size = _unpack_long(binary + pos[0])
    pos[0] += size + sizeof(INT64_t)


cdef inline void _pack_index(INT32_t index, bytearray buf):
    # 2 bytes
    _pack_short(index, buf)


cdef inline INT32_t _unpack_index(const char *binary, size_t* pos) nogil:
    cdef INT32_t ret = _unpack_short(binary + pos[0])
    pos[0] += 2
    return ret


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef inline void _skip_index(const char *binary, size_t* pos) nogil:
    pos[0] += 2


cdef inline void _pack_actor_ref(ActorRef actor_ref, bytearray buf):
    # this line circumvents a bug in Cython which led to application crash
    assert actor_ref is not None
    _pack_object(actor_ref.address, buf)
    _pack_object(actor_ref.uid, buf)


cdef inline ActorRef _unpack_actor_ref(const char *binary, size_t* pos):
    cdef object address
    cdef object uid

    address = _unpack_object(binary, pos)
    uid = _unpack_object(binary, pos)
    return ActorRef(address, uid)


cpdef bytes pack_actor_ref(ActorRef actor_ref):
    cdef bytearray buf = bytearray()

    _pack_actor_ref(actor_ref, buf)
    return bytes(buf)


cpdef void unpack_actor_ref(bytes binary, ActorRef actor_ref):
    cdef size_t pos = 0
    cdef object address
    cdef object uid

    address = _unpack_object(binary, &pos)
    uid = _unpack_object(binary, &pos)

    actor_ref.address = address
    actor_ref.uid = uid


cdef inline void _skip_address(const char *binary, size_t* pos) nogil:
    _skip_object(binary, pos)


cdef inline object _unpack_uid(const char *binary, size_t* pos):
    return _unpack_object(binary, pos)


cdef inline void _pack_message_size(size_t size, bytearray buf):
    _pack_long(size, buf)


cdef inline INT32_t _unpack_message_size(const char *binary, size_t* pos) nogil:
    cdef INT32_t ret = _unpack_long(binary + pos[0])
    pos[0] += 8
    return ret


@cython.nonecheck(False)
cdef inline list _pack_tuple_message(tuple messages):
    cdef list ret
    cdef bytearray bio
    cdef BYTE_t st
    cdef bytes m

    ret = []
    for message in messages:
        bio = bytearray()

        # use 1 byte to record the serialize type
        if isinstance(message, bytes):
            st = RAW_BYTES
            m = message
        else:
            st = PICKLE
            m = dumps(message)

        _pack_byte(st, bio)
        _pack_long(len(m), bio)
        ret.append(<bytes>bio)
        ret.append(m)

    return ret


cdef inline bytes _pack_sole_message(object message, bytearray buf):
    cdef BYTE_t st
    cdef bytes m

    # use 1 byte to record the serialize type
    if isinstance(message, bytes):
        st = RAW_BYTES
        m = message
    elif message is None:
        st = NONE
        m = bytes()
    else:
        st = PICKLE
        m = dumps(message)

    _pack_byte(st, buf)
    _pack_long(len(m), buf)
    return m


cdef inline list _pack_message(object message, bytearray buf):
    if isinstance(message, tuple):
        _pack_message_size(len(message), buf)
        return _pack_tuple_message(message)
    else:
        _pack_message_size(0, buf)
        return [_pack_sole_message(message, buf)]


cdef inline tuple _unpack_tuple_message(const char *buf, size_t size, size_t* pos):
    return tuple(_unpack_sole_message(buf, pos) for _ in range(size))


cdef inline object _unpack_sole_message(const char *binary, size_t* pos):
    cdef BYTE_t st
    cdef size_t size
    cdef bytes ret_bytes

    st = binary[pos[0]]
    pos[0] += 1
    size = _unpack_long(binary + pos[0])
    pos[0] += 8

    ret_bytes = binary[pos[0]:pos[0] + size]
    pos[0] += size
    if st == RAW_BYTES:
        return ret_bytes
    elif st == NONE:
        return None
    else:
        return loads(ret_bytes)


cdef inline object _unpack_message(const char *buf, size_t* pos):
    cdef size_t size

    size = _unpack_message_size(buf, pos)
    if size > 0:
        return _unpack_tuple_message(buf, size, pos)
    else:
        return _unpack_sole_message(buf, pos)


cdef inline object _pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
                                      bint send, object write=None, INT32_t protocol=DEFAULT_PROTOCOL):
    cdef bytes message_id
    cdef INT32_t message_type
    cdef bytearray buf
    cdef object m

    # from_index -1 means main process, -2 means remote, other is the subprocess id
    message_id = new_message_id()
    message_type = SEND_ALL if send else TELL_ALL

    buf = bytearray()
    _pack_message_type(message_type, buf, protocol=protocol)
    _pack_message_id(message_id, buf)
    _pack_index(from_index, buf)
    _pack_index(to_index, buf)
    _pack_actor_ref(actor_ref, buf)
    m = _pack_message(message, buf)

    if write is not None:
        write(buf, *m)
        return message_id

    return [message_id, buf] + m


cdef inline object _unpack_send_message(bytes binary, bint send=True):
    cdef object message_type
    cdef bytes message_id
    cdef INT32_t from_index
    cdef INT32_t to_index
    cdef ActorRef actor_ref
    cdef object message
    cdef size_t pos = 0
    cdef const char *binary_ptr = binary

    _unpack_message_type_value(binary_ptr, &pos)
    message_type = SEND_ALL if send else TELL_ALL
    message_id = _unpack_message_id(binary_ptr, &pos)
    from_index = _unpack_index(binary_ptr, &pos)
    to_index = _unpack_index(binary_ptr, &pos)
    actor_ref = _unpack_actor_ref(binary_ptr, &pos)
    message = _unpack_message(binary_ptr, &pos)

    return SEND_MESSAGE(message_type=message_type, message_id=message_id,
                        from_index=from_index, to_index=to_index,
                        actor_ref=actor_ref, message=message)


cpdef object pack_send_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
                               object write=None, INT32_t protocol=DEFAULT_PROTOCOL):
    return _pack_send_message(from_index, to_index, actor_ref, message, True,
                              write=write, protocol=protocol)


cpdef object unpack_send_message(bytes message):
    return _unpack_send_message(message)


cpdef object pack_tell_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object message,
                               object write=None, INT32_t protocol=DEFAULT_PROTOCOL):
    return _pack_send_message(from_index, to_index, actor_ref, message, False,
                              write=write, protocol=protocol)


cpdef object unpack_tell_message(bytes message):
    return _unpack_send_message(message, send=False)


cpdef tuple get_index(bytes binary, object calc_from_uid):
    cdef object buf
    cdef INT32_t from_index
    cdef INT32_t to_index
    cdef object uid
    cdef size_t pos = 0
    cdef const char *binary_ptr = binary

    _skip_message_type(binary_ptr, &pos)
    _skip_message_id(binary_ptr, &pos)

    from_index = _unpack_index(binary_ptr, &pos)
    to_index = _unpack_index(binary_ptr, &pos)
    if to_index != -1 or calc_from_uid is None:
        return from_index, to_index

    _skip_address(binary_ptr, &pos)

    uid = _unpack_uid(binary_ptr, &pos)
    return from_index, calc_from_uid(uid)


cpdef object pack_create_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref, object actor_cls,
                                       tuple args, dict kw, object write=None, INT32_t protocol=DEFAULT_PROTOCOL):
    cdef bytes message_id
    cdef bytearray buf

    # from_index -1 means main process, -2 means remote, other is the subprocess id
    message_id = new_message_id()

    buf = bytearray()
    _pack_message_type(CREATE_ACTOR, buf, protocol=protocol)
    _pack_message_id(message_id, buf)
    _pack_index(from_index, buf)
    _pack_index(to_index, buf)
    _pack_actor_ref(actor_ref, buf)
    _pack_object(actor_cls, buf)
    _pack_object(args if args else None, buf)
    _pack_object(kw if kw else None, buf)

    if write is not None:
        write(buf)
        return message_id

    return message_id, buf


cpdef object unpack_create_actor_message(bytes binary):
    cdef object buf
    cdef object message_type
    cdef bytes message_id
    cdef INT32_t from_index
    cdef INT32_t to_index
    cdef ActorRef actor_ref
    cdef object actor_cls
    cdef tuple args
    cdef dict kw
    cdef size_t pos = 0
    cdef const char *binary_ptr = binary

    _unpack_message_type_value(binary_ptr, &pos)
    message_type = CREATE_ACTOR
    message_id = _unpack_message_id(binary_ptr, &pos)
    from_index = _unpack_index(binary_ptr, &pos)
    to_index = _unpack_index(binary_ptr, &pos)
    actor_ref = _unpack_actor_ref(binary_ptr, &pos)
    actor_cls = _unpack_object(binary_ptr, &pos)
    args = _unpack_object(binary_ptr, &pos) or tuple()
    kw = _unpack_object(binary_ptr, &pos) or dict()

    return CREATE_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
                                from_index=from_index, to_index=to_index,
                                actor_ref=actor_ref, actor_cls=actor_cls,
                                args=args, kwargs=kw)


cpdef object pack_destroy_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,
                                        object write=None, INT32_t protocol=DEFAULT_PROTOCOL):
    cdef bytes message_id
    cdef bytearray buf

    # from_index -1 means main process, -2 means remote, other is the subprocess id
    message_id = new_message_id()

    buf = bytearray()
    _pack_message_type(DESTROY_ACTOR, buf, protocol=protocol)
    _pack_message_id(message_id, buf)
    _pack_index(from_index, buf)
    _pack_index(to_index, buf)
    _pack_actor_ref(actor_ref, buf)

    if write is not None:
        write(buf)
        return message_id

    return message_id, buf


cpdef object unpack_destroy_actor_message(bytes binary):
    cdef object buf
    cdef object message_type
    cdef bytes message_id
    cdef INT32_t from_index
    cdef INT32_t to_index
    cdef ActorRef actor_ref
    cdef size_t pos = 0
    cdef const char *binary_ptr = binary

    message_type = _unpack_message_type(binary_ptr, &pos)
    message_id = _unpack_message_id(binary_ptr, &pos)
    from_index = _unpack_index(binary_ptr, &pos)
    to_index = _unpack_index(binary_ptr, &pos)
    actor_ref = _unpack_actor_ref(binary_ptr, &pos)

    return DESTROY_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
                                 from_index=from_index, to_index=to_index,
                                 actor_ref=actor_ref)


cpdef object pack_has_actor_message(INT32_t from_index, INT32_t to_index, ActorRef actor_ref,
                                    object write=None, INT32_t protocol=DEFAULT_PROTOCOL):
    cdef bytearray buf

    # from_index -1 means main process, -2 means remote, other is the subprocess id
    message_id = new_message_id()

    buf = bytearray()
    _pack_message_type(HAS_ACTOR, buf, protocol=protocol)
    _pack_message_id(message_id, buf)
    _pack_index(from_index, buf)
    _pack_index(to_index, buf)
    _pack_actor_ref(actor_ref, buf)

    if write is not None:
        write(buf)
        return message_id

    return message_id, buf


cpdef object unpack_has_actor_message(bytes binary):
    cdef object message_type
    cdef bytes message_id
    cdef INT32_t from_index
    cdef INT32_t to_index
    cdef ActorRef actor_ref
    cdef size_t pos = 0
    cdef const char *binary_ptr = binary

    message_type = _unpack_message_type(binary_ptr, &pos)
    message_id = _unpack_message_id(binary_ptr, &pos)
    from_index = _unpack_index(binary_ptr, &pos)
    to_index = _unpack_index(binary_ptr, &pos)
    actor_ref = _unpack_actor_ref(binary_ptr, &pos)

    return HAS_ACTOR_MESSAGE(message_type=message_type, message_id=message_id,
                             from_index=from_index, to_index=to_index,
                             actor_ref=actor_ref)


cpdef object pack_result_message(bytes message_id, INT32_t from_index, INT32_t to_index, object result,
                                 object write=None, INT32_t protocol=DEFAULT_PROTOCOL):
    cdef bytearray buf

    buf = bytearray()
    _pack_message_type(RESULT, buf, protocol=protocol)
    _pack_message_id(message_id, buf)
    _pack_index(from_index, buf)
    _pack_index(to_index, buf)
    _pack_object(result, buf)

    if write is not None:
        write(buf)
        return message_id

    return message_id, buf


cpdef object unpack_result_message(bytes binary, object from_index=None, object to_index=None):
    cdef object message_type
    cdef bytes message_id
    cdef object result
    cdef size_t pos = 0
    cdef const char *binary_ptr = binary

    _unpack_message_type_value(binary_ptr, &pos)
    message_type = MessageType.result
    message_id = _unpack_message_id(binary_ptr, &pos)
    if from_index is not None:
        _skip_index(binary_ptr, &pos)
    else:
        from_index = _unpack_index(binary_ptr, &pos)
    if to_index is not None:
        _skip_index(binary_ptr, &pos)
    else:
        to_index = _unpack_index(binary_ptr, &pos)
    result = _unpack_object(binary_ptr, &pos)

    return RESULT_MESSAGE(message_type=message_type, message_id=message_id,
                          from_index=from_index, to_index=to_index,
                          result=result)


cpdef object pack_error_message(bytes message_id, INT32_t from_index, INT32_t to_index,
                                object error_type, object error, object tb,
                                object write=None, INT32_t protocol=DEFAULT_PROTOCOL):
    cdef bytearray buf

    buf = bytearray()
    _pack_message_type(ERROR, buf, protocol=protocol)
    _pack_message_id(message_id, buf)
    _pack_index(from_index, buf)
    _pack_index(to_index, buf)
    _pack_object(error_type, buf)
    _pack_object(error, buf)
    _pack_object(tb, buf)

    if write is not None:
        write(buf)
        return message_id

    return message_id, buf


cpdef object unpack_error_message(bytes binary):
    cdef object message_type
    cdef bytes message_id
    cdef INT32_t from_index
    cdef INT32_t to_index
    cdef object error_type
    cdef object error
    cdef object tb
    cdef size_t pos = 0
    cdef const char *binary_ptr = binary

    message_type = _unpack_message_type(binary_ptr, &pos)
    message_id = _unpack_message_id(binary_ptr, &pos)
    from_index = _unpack_index(binary_ptr, &pos)
    to_index = _unpack_index(binary_ptr, &pos)
    error_type = _unpack_object(binary_ptr, &pos)
    error = _unpack_object(binary_ptr, &pos)
    tb = _unpack_object(binary_ptr, &pos)

    return ERROR_MESSAGE(message_type=message_type, message_id=message_id,
                         from_index=from_index, to_index=to_index,
                         error_type=error_type, error=error, traceback=tb)


cdef inline bytes _wrap_read_func(object read_func, size_t size):
    cdef bytes read_bytes

    try:
         read_bytes = read_func(size)
    except ConnectionResetError:
        raise BrokenPipeError('The remote server is closed')

    if len(read_bytes) == 0:
        raise BrokenPipeError('The remote server is closed')

    return read_bytes


cpdef bytes read_remote_message(object read_func):
    cdef size_t size
    cdef bytearray buf
    cdef INT32_t received_size
    cdef INT32_t left
    cdef bytes read_bytes

    read_bytes = _wrap_read_func(read_func, 8)
    size = _unpack_long(read_bytes)
    buf = bytearray()
    received_size = 0
    left = size

    while True:
        read_bytes = _wrap_read_func(read_func, left)
        buf.extend(read_bytes)
        received_size += len(read_bytes)
        if received_size >= size:
            break
        left = size - received_size

    return bytes(buf)


def write_remote_message(write_func, *binary):
    cdef bytes size_bytes
    cdef INT64_t size = 0

    for b in binary:
        size += len(b)
    size_bytes = (<char *>&size)[:sizeof(INT64_t)]
    write_func(size_bytes)

    for b in binary:
        write_func(b)



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ... import tensor as mt
from ...operands import OutputType, OperandStage
from ...serialize import KeyField, Int32Field
from ...tensor.array_utils import as_same_device, device
from ...tensor.core import TensorOrder
from ...tensor.random import RandomStateField
from ...tiles import TilesError
from ...utils import recursive_tile, check_chunks_unknown_shape
from ..metrics import euclidean_distances
from ..operands import LearnOperand, LearnOperandMixin


def _kmeans_plus_plus_init(X,
                           x_squared_norms,
                           random_state,
                           n_clusters: int,
                           n_local_trials: int = None):
    n_samples, n_features = X.shape

    centers = mt.empty((n_clusters, n_features), dtype=X.dtype)

    assert x_squared_norms is not None, 'x_squared_norms None in _k_init'

    # Set the number of local seeding trials if none is given
    if n_local_trials is None:
        # This is what Arthur/Vassilvitskii tried, but did not report
        # specific results for other than mentioning in the conclusion
        # that it helped.
        n_local_trials = 2 + int(np.log(n_clusters))

    # Pick first center randomly
    center_id = random_state.randint(n_samples)
    if X.issparse():  # pragma: no cover
        centers[0] = X[center_id].todense()
    else:
        centers[0] = X[center_id]

    # Initialize list of closest distances and calculate current potential
    closest_dist_sq = euclidean_distances(
        centers[0, mt.newaxis], X, Y_norm_squared=x_squared_norms,
        squared=True)
    current_pot = closest_dist_sq.sum()

    # Pick the remaining n_clusters-1 points
    for c in range(1, n_clusters):
        # Choose center candidates by sampling with probability proportional
        # to the squared distance to the closest existing center
        rand_vals = random_state.random_sample(n_local_trials) * current_pot
        candidate_ids = mt.searchsorted(closest_dist_sq.cumsum(),
                                        rand_vals)
        # XXX: numerical imprecision can result in a candidate_id out of range
        candidate_ids = mt.clip(candidate_ids, None, closest_dist_sq.size - 1)

        # Compute distances to center candidates
        distance_to_candidates = euclidean_distances(
            X[candidate_ids], X, Y_norm_squared=x_squared_norms, squared=True)

        # update closest distances squared and potential for each candidate
        distance_to_candidates = mt.minimum(closest_dist_sq, distance_to_candidates)

        candidates_pot = distance_to_candidates.sum(axis=1)

        # Decide which candidate is the best
        best_candidate = mt.argmin(candidates_pot)
        current_pot = candidates_pot[best_candidate]
        closest_dist_sq = distance_to_candidates[best_candidate]
        best_candidate = candidate_ids[best_candidate]

        # Permanently add best center candidate found in local tries
        if X.issparse():  # pragma: no cover
            c_center = X[best_candidate].todense()
        else:
            c_center = X[best_candidate]

        centers[c] = c_center

    return centers


class KMeansPlusPlusInit(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_PLUS_PLUS_INIT

    _x = KeyField('x')
    _n_clusters = Int32Field('n_clusters')
    _x_squared_norms = KeyField('x_squared_norms')
    _state = RandomStateField('state')
    _n_local_trials = Int32Field('n_local_trials')

    def __init__(self, x=None, n_clusters=None, x_squared_norms=None,
                 state=None, n_local_trials=None, output_types=None, **kw):
        super().__init__(_x=x, _n_clusters=n_clusters, _x_squared_norms=x_squared_norms,
                         _state=state, _n_local_trials=n_local_trials,
                         _output_types=output_types, **kw)
        if self._output_types is None:
            self._output_types = [OutputType.tensor]

    @property
    def x(self):
        return self._x

    @property
    def n_clusters(self):
        return self._n_clusters

    @property
    def x_squared_norms(self):
        return self._x_squared_norms

    @property
    def state(self):
        return self._state

    @property
    def n_local_trials(self):
        return self._n_local_trials

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._x = self._inputs[0]
        self._x_squared_norms = self._inputs[-1]

    def __call__(self):
        inputs = [self._x, self._x_squared_norms]
        kw = {
            'shape': (self._n_clusters, self._x.shape[1]),
            'dtype': self._x.dtype,
            'order': TensorOrder.C_ORDER
        }
        return self.new_tileable(inputs, kws=[kw])

    @classmethod
    def _tile_one_chunk(cls, op: "KMeansPlusPlusInit"):
        out = op.outputs[0]

        chunk_op = op.copy().reset_key()
        chunk_kw = out.params.copy()
        chunk_kw['index'] = (0, 0)
        chunk_inputs = [op.x.chunks[0], op.x_squared_norms.chunks[0]]
        chunk = chunk_op.new_chunk(chunk_inputs, kws=[chunk_kw])

        kw = out.params
        kw['chunks'] = [chunk]
        kw['nsplits'] = tuple((s,) for s in out.shape)
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[kw])

    @classmethod
    def tile(cls, op: "KMeansPlusPlusInit"):
        if len(op.x.chunks) == 1:
            assert len(op.x_squared_norms.chunks) == 1
            return cls._tile_one_chunk(op)
        else:
            return cls._tile_k_init(op)

    @classmethod
    def _tile_k_init(cls, op: "KMeansPlusPlusInit"):
        X = op.x
        n_clusters = op.n_clusters
        x_squared_norms = op.x_squared_norms
        random_state = op.state
        n_local_trials = op.n_local_trials

        centers = _kmeans_plus_plus_init(X, x_squared_norms, random_state,
                                         n_clusters, n_local_trials)
        return recursive_tile(centers)

    @classmethod
    def execute(cls, ctx, op: "KMeansPlusPlusInit"):
        try:
            from sklearn.cluster._kmeans import _kmeans_plusplus
        except ImportError:  # pragma: no cover
            try:
                from sklearn.cluster._kmeans import _k_init
            except ImportError:
                from sklearn.cluster.k_means_ import _k_init

            def _kmeans_plusplus(*args, **kwargs):
                return _k_init(*args, **kwargs), None

        (x, x_squared_norms), device_id, _ = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True)

        with device(device_id):
            ctx[op.outputs[0].key] = _kmeans_plusplus(
                x, op.n_clusters, x_squared_norms=x_squared_norms, random_state=op.state,
                n_local_trials=op.n_local_trials)[0]


###############################################################################
# Initialization heuristic


def _k_init(X, n_clusters, x_squared_norms, random_state, n_local_trials=None):
    """Init n_clusters seeds according to k-means++

    Parameters
    ----------
    X : array or sparse matrix, shape (n_samples, n_features)
        The data to pick seeds for. To avoid memory copy, the input data
        should be double precision (dtype=np.float64).

    n_clusters : integer
        The number of seeds to choose

    x_squared_norms : array, shape (n_samples,)
        Squared Euclidean norm of each data point.

    random_state : int, RandomState instance
        The generator used to initialize the centers. Use an int to make the
        randomness deterministic.
        See :term:`Glossary <random_state>`.

    n_local_trials : integer, optional
        The number of seeding trials for each center (except the first),
        of which the one reducing inertia the most is greedily chosen.
        Set to None to make the number of trials depend logarithmically
        on the number of seeds (2+log(k)); this is the default.

    Notes
    -----
    Selects initial cluster centers for k-mean clustering in a smart way
    to speed up convergence. see: Arthur, D. and Vassilvitskii, S.
    "k-means++: the advantages of careful seeding". ACM-SIAM symposium
    on Discrete algorithms. 2007

    Version ported from http://www.stanford.edu/~darthur/kMeansppTest.zip,
    which is the implementation used in the aforementioned paper.
    """
    op = KMeansPlusPlusInit(x=X, n_clusters=n_clusters, x_squared_norms=x_squared_norms,
                            state=random_state, n_local_trials=n_local_trials)
    return op()


class KMeansScalablePlusPlusInit(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_SCALABLE_PLUS_PLUS_INIT

    _x = KeyField('x')
    _n_clusters = Int32Field('n_clusters')
    _x_squared_norms = KeyField('x_squared_norms')
    _state = RandomStateField('state')
    _init_iter = Int32Field('init_iter')
    _oversampling_factor = Int32Field('oversampling_factor')

    def __init__(self, x=None, n_clusters=None, x_squared_norms=None,
                 state=None, init_iter=None, oversampling_factor=None,
                 output_types=None, stage=None, **kw):
        super().__init__(_x=x, _n_clusters=n_clusters, _x_squared_norms=x_squared_norms,
                         _state=state, _init_iter=init_iter,
                         _oversampling_factor=oversampling_factor,
                         _stage=stage, _output_types=output_types, **kw)
        if self._output_types is None:
            self._output_types = [OutputType.tensor]

    @property
    def x(self):
        return self._x

    @property
    def n_clusters(self):
        return self._n_clusters

    @property
    def x_squared_norms(self):
        return self._x_squared_norms

    @property
    def state(self):
        return self._state

    @property
    def init_iter(self):
        return self._init_iter

    @property
    def oversampling_factor(self):
        return self._oversampling_factor

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self._x is not None:
            self._x = self._inputs[0]
        if self._x_squared_norms is not None:
            self._x_squared_norms = self._inputs[-1]

    def __call__(self):
        inputs = [self._x, self._x_squared_norms]
        kw = {
            'shape': (self._n_clusters, self._x.shape[1]),
            'dtype': self._x.dtype,
            'order': TensorOrder.C_ORDER
        }
        return self.new_tileable(inputs, kws=[kw])

    @classmethod
    def tile(cls, op: "KMeansScalablePlusPlusInit"):
        check_chunks_unknown_shape(op.inputs, TilesError)

        x = mt.tensor(op.x)
        x_squared_norms = mt.atleast_2d(op.x_squared_norms)
        out = op.outputs[0]

        random_state = op.state
        rs = mt.random.RandomState.from_numpy(random_state)

        n_samples, n_features = x.shape
        n_clusters = op.n_clusters

        # step 1, sample a centroid
        centers = x[random_state.randint(n_samples, size=1)]

        for _ in range(op.init_iter):
            distances = euclidean_distances(
                x, centers, X_norm_squared=x_squared_norms, squared=True)

            # calculate the cost of data with respect to current centers
            cost = mt.sum(mt.min(distances, axis=1))

            # calculate the distribution to sample new centers
            distribution = mt.full(len(distances), 1 / len(distances))
            mt.true_divide(mt.min(distances, axis=1), cost,
                           where=cost != 0, out=distribution)

            # pick new centers
            new_centers_size = op.oversampling_factor * n_clusters
            new_centers = x[rs.choice(n_samples, new_centers_size, p=distribution)]

            centers = mt.concatenate([centers, new_centers])

        # rechunk centers into one chunk
        centers = recursive_tile(centers).rechunk(centers.shape)

        distances = recursive_tile(euclidean_distances(
            x, centers, X_norm_squared=x_squared_norms, squared=True))

        map_index_to_chunks = {}
        # calculate weight for each chunk
        for c in distances.chunks:
            map_chunk_op = KMeansScalablePlusPlusInit(stage=OperandStage.map)
            map_chunk_kw = {
                'shape': (len(centers),),
                'dtype': np.dtype(np.int64),
                'order': TensorOrder.C_ORDER,
                'index': c.index
            }
            map_chunk = map_chunk_op.new_chunk([c], kws=[map_chunk_kw])
            map_index_to_chunks[c.index] = map_chunk

        combine_chunks = []
        for i in range(distances.chunk_shape[0]):
            map_chunks = [map_index_to_chunks[i, j]
                          for j in range(distances.chunk_shape[1])]
            combine_chunk_op = KMeansScalablePlusPlusInit(stage=OperandStage.combine)
            combine_chunk_kw = {
                'shape': (len(centers),),
                'dtype': np.dtype(np.int64),
                'order': TensorOrder.C_ORDER,
                'index': (i,)
            }
            combine_chunk = combine_chunk_op.new_chunk(
                map_chunks, kws=[combine_chunk_kw])
            combine_chunks.append(combine_chunk)

        reduce_chunk_op = KMeansScalablePlusPlusInit(n_clusters=op.n_clusters,
                                                     state=random_state,
                                                     stage=OperandStage.reduce)
        reduce_chunk_kw = out.params
        reduce_chunk_kw['index'] = (0, 0)
        reduce_chunk = reduce_chunk_op.new_chunk([centers.chunks[0]] + combine_chunks,
                                                 kws=[reduce_chunk_kw])

        new_op = op.copy()
        kw = out.params
        kw['chunks'] = [reduce_chunk]
        kw['nsplits'] = tuple((s,) for s in out.shape)
        return new_op.new_tileables(op.inputs, kws=[kw])

    @classmethod
    def _execute_map(cls, ctx, op: "KMeansScalablePlusPlusInit"):
        distances = ctx[op.inputs[0].key]
        min_distance_ids = np.argmin(distances, axis=1)
        min_distances = distances[range(len(distances)), min_distance_ids]
        ctx[op.outputs[0].key] = (min_distances, min_distance_ids)

    @classmethod
    def _execute_combine(cls, ctx, op: "KMeansScalablePlusPlusInit"):
        out = op.outputs[0]
        all_distances, all_min_distance_ids = tuple(zip(*(ctx[inp.key] for inp in op.inputs)))
        distances = np.stack(all_distances).T
        min_distance_ids = np.stack(all_min_distance_ids).T

        combined_min_distance_id = np.argmin(distances, axis=1)
        min_distance_ids = min_distance_ids[range(len(distances)), combined_min_distance_id]
        count = np.bincount(min_distance_ids)
        result = np.zeros(out.shape[0], dtype=np.int64)
        result[:len(count)] = count
        ctx[out.key] = result

    @classmethod
    def _execute_reduce(cls, ctx, op: "KMeansScalablePlusPlusInit"):
        from sklearn.cluster import KMeans

        inputs = [ctx[inp.key] for inp in op.inputs]

        count = np.zeros(inputs[1].shape[0], dtype=np.int64)
        for inp in inputs[1:]:
            count += inp
        weight = count / count.sum()

        centers = inputs[0]

        kmeans = KMeans(n_clusters=op.n_clusters, n_init=1,
                        random_state=op.state)
        kmeans.fit(centers, sample_weight=weight)
        ctx[op.outputs[0].key] = kmeans.cluster_centers_

    @classmethod
    def execute(cls, ctx, op: "KMeansScalablePlusPlusInit"):
        if op.stage == OperandStage.map:
            return cls._execute_map(ctx, op)
        elif op.stage == OperandStage.combine:
            return cls._execute_combine(ctx, op)
        else:
            return cls._execute_reduce(ctx, op)


def _scalable_k_init(X, n_clusters, x_squared_norms, random_state,
                     oversampling_factor=2, init_iter=5):
    op = KMeansScalablePlusPlusInit(x=X, n_clusters=n_clusters,
                                    x_squared_norms=x_squared_norms,
                                    state=random_state, init_iter=init_iter,
                                    oversampling_factor=oversampling_factor)
    return op()



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ...serialize import KeyField, AnyField, BoolField, Int32Field
from ...core import Entity, Base
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape, ceildiv, recursive_tile
from ..operands import TensorOperand, TensorOperandMixin
from ..datasource import tensor as astensor
from ..array_utils import as_same_device, device
from ..core import Tensor, TENSOR_TYPE
from ..utils import decide_unify_split


class TensorFillDiagonal(TensorOperand, TensorOperandMixin):
    _op_type_ = OperandDef.FILL_DIAGONAL

    _input = KeyField('input')
    _val = AnyField('val')
    _wrap = BoolField('wrap')
    # used for chunk
    _k = Int32Field('k')

    def __init__(self, val=None, wrap=None, k=None, dtype=None, **kw):
        super().__init__(_val=val, _wrap=wrap, _k=k, _dtype=dtype, **kw)

    @property
    def input(self):
        return self._input

    @property
    def val(self):
        return self._val

    @property
    def wrap(self):
        return self._wrap

    @property
    def k(self):
        return self._k

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if len(self._inputs) == 2:
            self._val = self._inputs[1]

    def __call__(self, a, val=None):
        inputs = [a]
        if val is not None:
            inputs.append(val)
        return self.new_tensor(inputs, shape=a.shape, order=a.order)

    @staticmethod
    def _process_val(val, a, wrap):
        """
        given the `val`, `a`, `wrap` which are the arguments in `fill_diagonal`,
        do some preprocess on `val` includes:

        1. calculate the length to fill on diagonal, 2-d and n-d(n > 2)
           as well as that `wrap` is True and `a` is a tall matrix need to be considered.
        2. if val is a Tensor, rechunk it into one chunk.
        """

        from ..datasource import diag
        from ..base import tile

        is_val_tensor = isinstance(val, TENSOR_TYPE)

        if a.ndim == 2:
            if wrap and TensorFillDiagonal._is_tall(a):
                size = sum(diag(sub).shape[0] for sub
                           in TensorFillDiagonal._split_tall_matrix(a))
            else:
                size = diag(a).shape[0]
        else:
            # every dimension has same shape
            size = a.shape[0]

        repeat_method = tile if is_val_tensor else np.tile
        val_size = val.size
        if val_size < size:
            n = ceildiv(size, val_size)
            val = repeat_method(val, n)[:size]
        elif val_size > size:
            val = val[:size]

        if is_val_tensor and val.ndim > 0:
            val = recursive_tile(val)
            val = val.rechunk({0: val.size})

        return recursive_tile(val) if is_val_tensor else val

    @staticmethod
    def _gen_val(val, diag_idx, cum_sizes):
        """
        Given a tensor-level `val`, calculate the chunk-level `val`.
        Consider both the cases that `val` could be a tensor or ndarray.

        :param val: tensor-level `val`
        :diag_idx: chunk index on the diagonal direction
        :cum_sizes: accumulative chunk sizes on the diagonal direction
        """
        from .slice import TensorSlice

        if val.ndim == 0:
            if isinstance(val, TENSOR_TYPE):
                return val.chunks[0]
            else:
                return val

        if isinstance(val, TENSOR_TYPE):
            start, stop = cum_sizes[diag_idx], cum_sizes[diag_idx + 1]
            slc = slice(start, stop)
            slc_op = TensorSlice(slices=[slc], dtype=val.dtype)
            return slc_op.new_chunk([val.chunks[0]], shape=(stop - start,),
                                    order=val.order, index=(diag_idx,))
        else:
            return val[cum_sizes[diag_idx]: cum_sizes[diag_idx + 1]]

    @classmethod
    def _tile_2d(cls, op, val):
        from ..datasource import diag

        d = diag(op.input)._inplace_tile()
        index_to_diag_chunk = {c.inputs[0].index: c for c in d.chunks}
        cum_sizes = [0] + np.cumsum(d.nsplits[0]).tolist()

        out_chunks = []
        for chunk in op.input.chunks:
            if chunk.index not in index_to_diag_chunk:
                out_chunks.append(chunk)
            else:
                diag_chunk = index_to_diag_chunk[chunk.index]
                diag_idx = diag_chunk.index[0]
                input_chunks = [chunk]
                chunk_val = cls._gen_val(val, diag_idx, cum_sizes)
                if len(op.inputs) == 2:
                    input_chunks.append(chunk_val)
                chunk_op = op.copy().reset_key()
                chunk_op._wrap = False
                chunk_op._k = diag_chunk.op.k
                chunk_op._val = chunk_val
                out_chunk = chunk_op.new_chunk(input_chunks, shape=chunk.shape,
                                               order=chunk.order,
                                               index=chunk.index)
                out_chunks.append(out_chunk)

        out = op.outputs[0]
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out.shape, order=out.order,
                                  chunks=out_chunks, nsplits=op.input.nsplits)

    @classmethod
    def _tile_nd(cls, op, val):
        # if more than 3d, we will rechunk the tensor into square chunk
        # on the diagonal direction
        in_tensor = op.input
        nsplits = np.array(in_tensor.nsplits)
        if not np.issubdtype(nsplits.dtype, np.integer) or \
                not np.all(np.diff(nsplits, axis=1) == 0):
            # need rechunk
            nsplit = decide_unify_split(*in_tensor.nsplits)
            in_tensor = in_tensor.rechunk(
                tuple(nsplit for _ in range(in_tensor.ndim)))._inplace_tile()
        cum_sizes = [0] + np.cumsum(in_tensor.nsplits[0]).tolist()

        out_chunks = []
        for chunk in in_tensor.chunks:
            if len(set(chunk.index)) == 1:
                # chunk on the diagonal direction
                chunk_op = op.copy().reset_key()
                chunk_op._k = 0
                chunk_inputs = [chunk]
                chunk_val = cls._gen_val(val, chunk.index[0], cum_sizes)
                if len(op.inputs) == 2:
                    chunk_inputs.append(chunk_val)
                chunk_op._val = chunk_val
                out_chunk = chunk_op.new_chunk(chunk_inputs, shape=chunk.shape,
                                               order=chunk.order,
                                               index=chunk.index)
                out_chunks.append(out_chunk)
            else:
                out_chunks.append(chunk)

        out = op.outputs[0]
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out.shape, order=out.order,
                                  chunks=out_chunks, nsplits=in_tensor.nsplits)

    @classmethod
    def _tile_one_chunk(cls, op, val):
        out = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk_inputs = [op.input.chunks[0]]
        if isinstance(val, TENSOR_TYPE):
            chunk_inputs.append(val.chunks[0])
        chunk = chunk_op.new_chunk(chunk_inputs, shape=out.shape,
                                   order=out.order, index=(0,) * out.ndim)
        new_op = op.copy()
        return new_op.new_tensors(op.inputs, shape=out.shape, order=out.order,
                                  chunks=[chunk], nsplits=((s,) for s in out.shape))

    @staticmethod
    def _is_tall(x):
        return x.shape[0] > x.shape[1] + 1

    @staticmethod
    def _split_tall_matrix(a):
        blocksize = a.shape[1] + 1
        n_block = ceildiv(a.shape[0], blocksize)
        return [a[i * blocksize: (i + 1) * blocksize]
                for i in range(n_block)]

    @classmethod
    def tile(cls, op):
        # input tensor must have no unknown chunk shape
        check_chunks_unknown_shape(op.inputs, TilesError)

        in_tensor = op.input
        is_in_tensor_tall = cls._is_tall(in_tensor)

        if op.val.ndim > 0:
            val = cls._process_val(op.val, in_tensor, op.wrap)
        else:
            val = op.val

        if len(in_tensor.chunks) == 1:
            return cls._tile_one_chunk(op, val)

        if op.input.ndim == 2:
            if op.wrap and is_in_tensor_tall:
                from ..merge import concatenate

                sub_tensors = cls._split_tall_matrix(in_tensor)
                for i, sub_tensor in enumerate(sub_tensors):
                    if val.ndim > 0:
                        sub_val = val[i * sub_tensor.shape[1]:
                                      (i + 1) * sub_tensor.shape[1]]
                    else:
                        sub_val = val
                    fill_diagonal(sub_tensor, sub_val, wrap=False)
                out_tensor = concatenate(sub_tensors)
                return [recursive_tile(out_tensor)]
            else:
                return cls._tile_2d(op, val)
        else:
            return cls._tile_nd(op, val)

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True)
        a = inputs[0]
        if len(inputs) == 2:
            val = inputs[1]
        else:
            val = op.val

        with device(device_id):
            if not op.k:
                a = a.copy()
                xp.fill_diagonal(a, val, wrap=op.wrap)
            else:
                assert a.ndim == 2
                k = op.k or 0
                n_rows, n_cols = a.shape
                if k > 0:
                    n_cols -= k
                elif k < 0:
                    n_rows += k
                n = min(n_rows, n_cols)

                # generate indices
                rows, cols = np.diag_indices(n)
                if k > 0:
                    cols = cols.copy()
                    cols += k
                elif k < 0:
                    rows = rows.copy()
                    rows -= k

                a = a.copy()
                a[rows, cols] = val

            ctx[op.outputs[0].key] = a


def fill_diagonal(a, val, wrap=False):
    """Fill the main diagonal of the given tensor of any dimensionality.

    For a tensor `a` with ``a.ndim >= 2``, the diagonal is the list of
    locations with indices ``a[i, ..., i]`` all identical. This function
    modifies the input tensor in-place, it does not return a value.

    Parameters
    ----------
    a : Tensor, at least 2-D.
      Tensor whose diagonal is to be filled, it gets modified in-place.

    val : scalar
      Value to be written on the diagonal, its type must be compatible with
      that of the tensor a.

    wrap : bool
      For tall matrices in NumPy version up to 1.6.2, the
      diagonal "wrapped" after N columns. You can have this behavior
      with this option. This affects only tall matrices.

    See also
    --------
    diag_indices, diag_indices_from

    Notes
    -----

    This functionality can be obtained via `diag_indices`, but internally
    this version uses a much faster implementation that never constructs the
    indices and uses simple slicing.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = mt.zeros((3, 3), int)
    >>> mt.fill_diagonal(a, 5)
    >>> a.execute()
    array([[5, 0, 0],
           [0, 5, 0],
           [0, 0, 5]])

    The same function can operate on a 4-D tensor:

    >>> a = mt.zeros((3, 3, 3, 3), int)
    >>> mt.fill_diagonal(a, 4)

    We only show a few blocks for clarity:

    >>> a[0, 0].execute()
    array([[4, 0, 0],
           [0, 0, 0],
           [0, 0, 0]])
    >>> a[1, 1].execute()
    array([[0, 0, 0],
           [0, 4, 0],
           [0, 0, 0]])
    >>> a[2, 2].execute()
    array([[0, 0, 0],
           [0, 0, 0],
           [0, 0, 4]])

    The wrap option affects only tall matrices:

    >>> # tall matrices no wrap
    >>> a = mt.zeros((5, 3), int)
    >>> mt.fill_diagonal(a, 4)
    >>> a.execute()
    array([[4, 0, 0],
           [0, 4, 0],
           [0, 0, 4],
           [0, 0, 0],
           [0, 0, 0]])

    >>> # tall matrices wrap
    >>> a = mt.zeros((5, 3), int)
    >>> mt.fill_diagonal(a, 4, wrap=True)
    >>> a.execute()
    array([[4, 0, 0],
           [0, 4, 0],
           [0, 0, 4],
           [0, 0, 0],
           [4, 0, 0]])

    >>> # wide matrices
    >>> a = mt.zeros((3, 5), int)
    >>> mt.fill_diagonal(a, 4, wrap=True)
    >>> a.execute()
    array([[4, 0, 0, 0, 0],
           [0, 4, 0, 0, 0],
           [0, 0, 4, 0, 0]])

    The anti-diagonal can be filled by reversing the order of elements
    using either `numpy.flipud` or `numpy.fliplr`.

    >>> a = mt.zeros((3, 3), int)
    >>> mt.fill_diagonal(mt.fliplr(a), [1,2,3])  # Horizontal flip
    >>> a.execute()
    array([[0, 0, 1],
           [0, 2, 0],
           [3, 0, 0]])
    >>> mt.fill_diagonal(mt.flipud(a), [1,2,3])  # Vertical flip
    >>> a.execute()
    array([[0, 0, 3],
           [0, 2, 0],
           [1, 0, 0]])

    Note that the order in which the diagonal is filled varies depending
    on the flip function.
    """

    if not isinstance(a, Tensor):
        raise TypeError(f'`a` should be a tensor, got {type(a)}')
    if a.ndim < 2:
        raise ValueError("array must be at least 2-d")
    if a.ndim > 2 and len(set(a.shape)) != 1:
        raise ValueError("All dimensions of input must be of equal length")

    # process val
    if isinstance(val, (Base, Entity)):
        val = astensor(val)
        if val.ndim > 1:
            val = val.ravel()
        val_input = val
    else:
        val = np.asarray(val)
        if val.ndim > 1:
            val = val.ravel()
        val_input = None

    op = TensorFillDiagonal(val=val, wrap=wrap, dtype=a.dtype)
    t = op(a, val=val_input)
    a.data = t.data



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from numbers import Integral

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import Base, Entity, ExecutableTuple, OutputType
from ...context import get_context
from ...serialize import KeyField, AnyField, BoolField, Int32Field, StringField
from ...tensor import tensor as astensor
from ...tensor.core import TENSOR_TYPE, TensorOrder
from ...tiles import TilesError
from ...utils import check_chunks_unknown_shape
from ..core import SERIES_TYPE, INDEX_TYPE
from ..datasource.index import from_pandas as asindex
from ..initializer import Series as asseries
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index


class DataFrameCut(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.CUT

    _input = KeyField('input')
    _bins = AnyField('bins')
    _right = BoolField('right')
    _labels = AnyField('labels')
    _retbins = BoolField('retbins')
    _precision = Int32Field('precision')
    _include_lowest = BoolField('include_lowest')
    _duplicates = StringField('duplicates')
    _input_min = KeyField('input_min')
    _input_max = KeyField('input_max')

    def __init__(self, bins=None, right=None, labels=None, retbins=None,
                 precision=None, include_lowest=None, duplicates=None,
                 input_min=None, input_max=None, **kw):
        super().__init__(_bins=bins, _right=right, _labels=labels,
                         _retbins=retbins, _precision=precision,
                         _include_lowest=include_lowest, _duplicates=duplicates,
                         _input_min=input_min, _input_max=input_max, **kw)

    @property
    def input(self):
        return self._input

    @property
    def bins(self):
        return self._bins

    @property
    def right(self):
        return self._right

    @property
    def labels(self):
        return self._labels

    @property
    def retbins(self):
        return self._retbins

    @property
    def precision(self):
        return self._precision

    @property
    def include_lowest(self):
        return self._include_lowest

    @property
    def duplicates(self):
        return self._duplicates

    @property
    def input_min(self):
        return self._input_min

    @property
    def input_max(self):
        return self._input_max

    @property
    def output_limit(self):
        return 1 if not self._retbins else 2

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        self._input = next(inputs_iter)
        if self._input_min is not None:
            self._input_min = next(inputs_iter)
        if self._input_max is not None:
            self._input_max = next(inputs_iter)
        if isinstance(self._bins, (Base, Entity)):
            self._bins = next(inputs_iter)
        if isinstance(self._labels, (Base, Entity)):
            self._labels = next(inputs_iter)

    def __call__(self, x):
        if isinstance(x, pd.Series):
            x = asseries(x)
        elif not isinstance(x, (Base, Entity)):
            x = astensor(x)
        if x.ndim != 1:
            raise ValueError('Input array must be 1 dimensional')
        if x.size == 0:
            raise ValueError('Cannot cut empty array')

        inputs = [x]
        if isinstance(self.bins, Integral):
            self._input_min = input_min = x.min()
            self._input_max = input_max = x.max()
            inputs.extend([input_min, input_max])

        if self._labels is not None and \
                not isinstance(self._labels, (bool, Base, Entity)):
            self._labels = np.asarray(self._labels)

        # infer dtype
        x_empty = pd.Series([1], dtype=x.dtype) if isinstance(x, SERIES_TYPE) else \
            np.asarray([1], dtype=x.dtype)
        if isinstance(self._bins, INDEX_TYPE):
            bins = self._bins.index_value.to_pandas()
            inputs.append(self._bins)
            bins_unknown = True
        elif isinstance(self._bins, (Base, Entity)):
            bins = np.asarray([2], dtype=self._bins.dtype)
            inputs.append(self._bins)
            bins_unknown = True
        else:
            bins = self._bins
            bins_unknown = isinstance(self._bins, Integral)
        if isinstance(self._labels, (Base, Entity)):
            bins_unknown = True
            labels = None
            inputs.append(self._labels)
        else:
            if self._labels is False or not bins_unknown:
                labels = self._labels
            else:
                labels = None
        ret = pd.cut(x_empty, bins, right=self._right, labels=labels,
                     retbins=True, include_lowest=self._include_lowest,
                     duplicates=self._duplicates)

        kws = []
        output_types = []
        if bins_unknown and isinstance(ret[0].dtype, pd.CategoricalDtype):
            # inaccurate dtype, just create an empty one
            out_dtype = pd.CategoricalDtype()
        else:
            out_dtype = ret[0].dtype
        if isinstance(ret[0], pd.Series):
            output_types.append(OutputType.series)
            kws.append({
                'dtype': out_dtype,
                'shape': x.shape,
                'index_value': x.index_value,
                'name': x.name
            })
        elif isinstance(ret[0], np.ndarray):
            output_types.append(OutputType.tensor)
            kws.append({
                'dtype': out_dtype,
                'shape': x.shape,
                'order': TensorOrder.C_ORDER
            })
        else:
            assert isinstance(ret[0], pd.Categorical)
            output_types.append(OutputType.categorical)
            kws.append({
                'dtype': out_dtype,
                'shape': x.shape,
                'categories_value': parse_index(out_dtype.categories,
                                                store_data=True)
            })

        if self._retbins:
            if isinstance(self._bins, (pd.IntervalIndex, INDEX_TYPE)):
                output_types.append(OutputType.index)
                kws.append({
                    'dtype': self._bins.dtype,
                    'shape': self._bins.shape,
                    'index_value': self._bins.index_value
                        if isinstance(self._bins, INDEX_TYPE) else
                        parse_index(self._bins, store_data=False),
                    'name': self._bins.name
                })
            else:
                output_types.append(OutputType.tensor)
                kws.append({
                    'dtype': ret[1].dtype,
                    'shape': ret[1].shape if ret[1].size > 0 else (np.nan,),
                    'order': TensorOrder.C_ORDER
                })

        self.output_types = output_types
        return ExecutableTuple(self.new_tileables(inputs, kws=kws))

    @classmethod
    def tile(cls, op):
        if isinstance(op.bins, (Base, Entity)):
            # check op.bins chunk shapes
            check_chunks_unknown_shape([op.bins], TilesError)
            bins = op.bins.rechunk(op.bins.shape)._inplace_tile()
        else:
            bins = op.bins

        if isinstance(op.labels, (Base, Entity)):
            # check op.labels chunk shapes
            check_chunks_unknown_shape([op.labels], TilesError)
            labels = op.labels.rechunk(op.labels.shape)._inplace_tile()
        else:
            labels = op.labels

        if op.input_min is not None:
            # check if input_min and input_max executed
            input_min_chunk = op.input_min.chunks[0]
            input_max_chunk = op.input_max.chunks[0]
            ctx = get_context()
            keys = [input_min_chunk.key, input_max_chunk.key]
            metas = ctx.get_chunk_metas(keys)
            if any(meta is None for meta in metas):
                # not executed before
                raise TilesError('min and max of x have not been executed yet')

            # get min and max of x
            min_val, max_val = ctx.get_chunk_results(keys)
            # calculate bins
            if np.isinf(min_val) or np.isinf(max_val):
                raise ValueError('cannot specify integer `bins` '
                                 'when input data contains infinity')
            elif min_val == max_val:  # adjust end points before binning
                min_val -= 0.001 * abs(min_val) if min_val != 0 else 0.001
                max_val += 0.001 * abs(max_val) if max_val != 0 else 0.001
                bins = np.linspace(min_val, max_val, bins + 1, endpoint=True)
            else:  # adjust end points before binning
                bins = np.linspace(min_val, max_val, bins + 1, endpoint=True)
                adj = (max_val - min_val) * 0.001  # 0.1% of the range
                if op.right:
                    bins[0] -= adj
                else:
                    bins[-1] += adj

        outs = op.outputs

        out_chunks = []
        for c in op.input.chunks:
            chunk_op = op.copy().reset_key()
            chunk_inputs = [c]
            # make sure input_min and input_max reset
            chunk_op._input_min = chunk_op._input_max = None
            chunk_op._bins = bins
            # do not return bins always for chunk
            chunk_op._retbins = False
            if isinstance(bins, (Base, Entity)):
                chunk_inputs.append(bins.chunks[0])
            chunk_op._labels = labels
            if isinstance(labels, (Base, Entity)):
                chunk_inputs.append(labels.chunks[0])

            chunk_kws = []
            if isinstance(outs[0], SERIES_TYPE):
                chunk_kws.append({
                    'dtype': outs[0].dtype,
                    'shape': c.shape,
                    'index_value': c.index_value,
                    'name': c.name,
                    'index': c.index,
                })
            elif isinstance(outs[0], TENSOR_TYPE):
                chunk_kws.append({
                    'dtype': outs[0].dtype,
                    'shape': c.shape,
                    'order': TensorOrder.C_ORDER,
                    'index': c.index,
                })
            else:
                chunk_kws.append({
                    'dtype': outs[0].dtype,
                    'shape': c.shape,
                    'categories_value': outs[0].categories_value,
                    'index': c.index,
                })

            out_chunks.append(chunk_op.new_chunk(chunk_inputs, kws=chunk_kws))

        kws = []
        out_kw = outs[0].params
        out_kw['chunks'] = out_chunks
        out_kw['nsplits'] = op.input.nsplits
        kws.append(out_kw)
        if len(outs) == 2:
            bins_kw = outs[1].params
            bins_kw['chunks'] = bins_chunks = []
            if isinstance(bins, (Base, Entity)):
                bins_chunks.append(bins.chunks[0])
            else:
                if op.duplicates == 'drop':
                    if isinstance(bins, (np.ndarray, list, tuple)):
                        bins = np.unique(bins)
                    else:
                        bins = bins.unique()
                    bins = bins.astype(outs[1].dtype, copy=False)
                convert = \
                    astensor if not isinstance(bins, pd.IntervalIndex) else asindex
                bins_chunks.append(
                    convert(bins, chunk_size=len(bins))._inplace_tile().chunks[0])
            bins_kw['nsplits'] = ((len(bins),),)
            kws.append(bins_kw)
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=kws)

    @classmethod
    def execute(cls, ctx, op):
        x = ctx[op.input.key]
        bins = ctx[op.bins.key] if isinstance(op.bins, (Base, Entity)) else op.bins
        labels = ctx[op.labels.key] if isinstance(op.labels, (Base, Entity)) else op.labels

        cut = partial(pd.cut, right=op.right, retbins=op.retbins, precision=op.precision,
                      include_lowest=op.include_lowest, duplicates=op.duplicates)
        try:
            ret = cut(x, bins, labels=labels)
        except ValueError:
            # fail due to buffer source array is read-only
            ret = cut(x.copy(), bins, labels=labels)
        if op.retbins:  # pragma: no cover
            ctx[op.outputs[0].key] = ret[0]
            ctx[op.outputs[1].key] = ret[1]
        else:
            ctx[op.outputs[0].key] = ret


def cut(x, bins, right: bool = True, labels=None, retbins: bool = False,
        precision: int = 3, include_lowest: bool = False, duplicates: str = 'raise'):
    """
    Bin values into discrete intervals.

    Use `cut` when you need to segment and sort data values into bins. This
    function is also useful for going from a continuous variable to a
    categorical variable. For example, `cut` could convert ages to groups of
    age ranges. Supports binning into an equal number of bins, or a
    pre-specified array of bins.

    Parameters
    ----------
    x : array-like
        The input array to be binned. Must be 1-dimensional.
    bins : int, sequence of scalars, or IntervalIndex
        The criteria to bin by.

        * int : Defines the number of equal-width bins in the range of `x`. The
          range of `x` is extended by .1% on each side to include the minimum
          and maximum values of `x`.
        * sequence of scalars : Defines the bin edges allowing for non-uniform
          width. No extension of the range of `x` is done.
        * IntervalIndex : Defines the exact bins to be used. Note that
          IntervalIndex for `bins` must be non-overlapping.

    right : bool, default True
        Indicates whether `bins` includes the rightmost edge or not. If
        ``right == True`` (the default), then the `bins` ``[1, 2, 3, 4]``
        indicate (1,2], (2,3], (3,4]. This argument is ignored when
        `bins` is an IntervalIndex.
    labels : array or False, default None
        Specifies the labels for the returned bins. Must be the same length as
        the resulting bins. If False, returns only integer indicators of the
        bins. This affects the type of the output container (see below).
        This argument is ignored when `bins` is an IntervalIndex. If True,
        raises an error.
    retbins : bool, default False
        Whether to return the bins or not. Useful when bins is provided
        as a scalar.
    precision : int, default 3
        The precision at which to store and display the bins labels.
    include_lowest : bool, default False
        Whether the first interval should be left-inclusive or not.
    duplicates : {default 'raise', 'drop'}, optional
        If bin edges are not unique, raise ValueError or drop non-uniques.

    Returns
    -------
    out : Categorical, Series, or Tensor
        An array-like object representing the respective bin for each value
        of `x`. The type depends on the value of `labels`.

        * True (default) : returns a Series for Series `x` or a
          Categorical for all other inputs. The values stored within
          are Interval dtype.

        * sequence of scalars : returns a Series for Series `x` or a
          Categorical for all other inputs. The values stored within
          are whatever the type in the sequence is.

        * False : returns a tensor of integers.

    bins : Tensor or IntervalIndex.
        The computed or specified bins. Only returned when `retbins=True`.
        For scalar or sequence `bins`, this is a tensor with the computed
        bins. If set `duplicates=drop`, `bins` will drop non-unique bin. For
        an IntervalIndex `bins`, this is equal to `bins`.

    See Also
    --------
    qcut : Discretize variable into equal-sized buckets based on rank
        or based on sample quantiles.
    Categorical : Array type for storing data that come from a
        fixed set of values.
    Series : One-dimensional array with axis labels (including time series).
    IntervalIndex : Immutable Index implementing an ordered, sliceable set.

    Notes
    -----
    Any NA values will be NA in the result. Out of bounds values will be NA in
    the resulting Series or Categorical object.

    Examples
    --------
    Discretize into three equal-sized bins.

    >>> import mars.tensor as mt
    >>> import mars.dataframe as md

    >>> md.cut(mt.array([1, 7, 5, 4, 6, 3]), 3).execute()
    ... # doctest: +ELLIPSIS
    [(0.994, 3.0], (5.0, 7.0], (3.0, 5.0], (3.0, 5.0], (5.0, 7.0], ...
    Categories (3, interval[float64]): [(0.994, 3.0] < (3.0, 5.0] ...

    >>> md.cut(mt.array([1, 7, 5, 4, 6, 3]), 3, retbins=True).execute()
    ... # doctest: +ELLIPSIS
    ([(0.994, 3.0], (5.0, 7.0], (3.0, 5.0], (3.0, 5.0], (5.0, 7.0], ...
    Categories (3, interval[float64]): [(0.994, 3.0] < (3.0, 5.0] ...
    array([0.994, 3.   , 5.   , 7.   ]))

    Discovers the same bins, but assign them specific labels. Notice that
    the returned Categorical's categories are `labels` and is ordered.

    >>> md.cut(mt.array([1, 7, 5, 4, 6, 3]),
    ...        3, labels=["bad", "medium", "good"]).execute()
    [bad, good, medium, medium, good, bad]
    Categories (3, object): [bad < medium < good]

    ``labels=False`` implies you just want the bins back.

    >>> md.cut([0, 1, 1, 2], bins=4, labels=False).execute()
    array([0, 1, 1, 3])

    Passing a Series as an input returns a Series with categorical dtype:

    >>> s = md.Series(mt.array([2, 4, 6, 8, 10]),
    ...               index=['a', 'b', 'c', 'd', 'e'])
    >>> md.cut(s, 3).execute()
    ... # doctest: +ELLIPSIS
    a    (1.992, 4.667]
    b    (1.992, 4.667]
    c    (4.667, 7.333]
    d     (7.333, 10.0]
    e     (7.333, 10.0]
    dtype: category
    Categories (3, interval[float64]): [(1.992, 4.667] < (4.667, ...

    Passing a Series as an input returns a Series with mapping value.
    It is used to map numerically to intervals based on bins.

    >>> s = md.Series(mt.array([2, 4, 6, 8, 10]),
    ...               index=['a', 'b', 'c', 'd', 'e'])
    >>> md.cut(s, [0, 2, 4, 6, 8, 10], labels=False, retbins=True, right=False).execute()
    ... # doctest: +ELLIPSIS
    (a    0.0
     b    1.0
     c    2.0
     d    3.0
     e    NaN
     dtype: float64, array([0, 2, 4, 6, 8, 10]))

    Use `drop` optional when bins is not unique

    >>> md.cut(s, [0, 2, 4, 6, 10, 10], labels=False, retbins=True,
    ...        right=False, duplicates='drop').execute()
    ... # doctest: +ELLIPSIS
    (a    0.0
     b    1.0
     c    2.0
     d    3.0
     e    NaN
     dtype: float64, array([0, 2, 4, 6, 10]))

    Passing an IntervalIndex for `bins` results in those categories exactly.
    Notice that values not covered by the IntervalIndex are set to NaN. 0
    is to the left of the first bin (which is closed on the right), and 1.5
    falls between two bins.

    >>> bins = md.Index(pd.IntervalIndex.from_tuples([(0, 1), (2, 3), (4, 5)]))
    >>> md.cut([0, 0.5, 1.5, 2.5, 4.5], bins).execute()
    [NaN, (0, 1], NaN, (2, 3], (4, 5]]
    Categories (3, interval[int64]): [(0, 1] < (2, 3] < (4, 5]]
    """

    if isinstance(bins, Integral) and bins < 1:
        raise ValueError('`bins` should be a positive integer')

    op = DataFrameCut(bins=bins, right=right, labels=labels,
                      retbins=retbins, precision=precision,
                      include_lowest=include_lowest, duplicates=duplicates)
    ret = op(x)
    if not retbins:
        return ret[0]
    else:
        return ret
